/*
 * Support for Intel AES-NI instructions. This file contains glue
 * code, the real AES implementation is in intel-aes_asm.S.
 *
 * Copyright (C) 2008, Intel Corp.
 *    Author: Huang Ying <ying.huang@intel.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <linux/hardirq.h>
#include <asm/i387.h>
#include <asm/aes.h>

#define INTEL_AES_ALIGN	16

#define AES_MIN_KEY_SIZE        16
#define AES_MAX_KEY_SIZE        32

#define AES_BLOCK_SIZE          16

struct aes_ctx {
	u32 key_enc[60];
	u32 key_dec[60];
	u32 key_length;
};

asmlinkage int intel_aes_set_key(struct aes_ctx *ctx, const u8 *in_key,
				unsigned int key_len);
asmlinkage void intel_aes_enc(struct aes_ctx *ctx, u8 *out, const u8 *in);
asmlinkage void intel_aes_dec(struct aes_ctx *ctx, u8 *out, const u8 *in);

static inline struct aes_ctx *aes_ctx(struct crypto_tfm *tfm)
{
	unsigned long addr = (unsigned long)crypto_tfm_ctx(tfm);
	unsigned long align = INTEL_AES_ALIGN;

	if (align <= crypto_tfm_ctx_alignment())
		align = 1;
	return (struct aes_ctx *)ALIGN(addr, align);
}

static int aes_set_key(struct crypto_tfm *tfm, const u8 *in_key,
		       unsigned int key_len)
{
	struct aes_ctx *ctx = aes_ctx(tfm);
	u32 *tfm_flags = &tfm->crt_flags;
	int ret;

	if (key_len % 8) {
		*tfm_flags |= CRYPTO_TFM_RES_BAD_KEY_LEN;
		return -EINVAL;
	}

	if (in_interrupt())
		ret = crypto_aes_set_key_x86(tfm, in_key, key_len, tfm_flags);
	else {
		kernel_fpu_begin();
		ret = intel_aes_set_key(ctx, in_key, key_len);
		kernel_fpu_end();
	}

	return ret;
}

static void aes_encrypt(struct crypto_tfm *tfm, u8 *dst, const u8 *src)
{
	struct aes_ctx *ctx = aes_ctx(tfm);

	if (in_interrupt())
		crypto_aes_encrypt_x86(tfm, dst, src);
	else {
		kernel_fpu_begin();
		intel_aes_enc(ctx, dst, src);
		kernel_fpu_end();
	}
}

static void aes_decrypt(struct crypto_tfm *tfm, u8 *dst, const u8 *src)
{
	struct aes_ctx *ctx = aes_ctx(tfm);

	if (in_interrupt())
		crypto_aes_decrypt_x86(tfm, dst, src);
	else {
		kernel_fpu_begin();
		intel_aes_dec(ctx, dst, src);
		kernel_fpu_end();
	}
}

static struct crypto_alg intel_aes_alg = {
	.cra_name		= "aes",
	.cra_driver_name	= "aes-intel",
	.cra_priority		= 300,
	.cra_flags		= CRYPTO_ALG_TYPE_CIPHER,
	.cra_blocksize		= AES_BLOCK_SIZE,
	.cra_ctxsize		= sizeof(struct aes_ctx),
	.cra_alignmask		= INTEL_AES_ALIGN - 1,
	.cra_module		= THIS_MODULE,
	.cra_list		= LIST_HEAD_INIT(intel_aes_alg.cra_list),
	.cra_u	= {
		.cipher	= {
			.cia_min_keysize	= AES_MIN_KEY_SIZE,
			.cia_max_keysize	= AES_MAX_KEY_SIZE,
			.cia_setkey		= aes_set_key,
			.cia_encrypt		= aes_encrypt,
			.cia_decrypt		= aes_decrypt
		}
	}
};

static int __init intel_aes_init(void)
{
	if (!cpu_has_aes) {
		printk(KERN_ERR "Intel AES instructions are not detected.\n");
		return -ENODEV;
	}
	return crypto_register_alg(&intel_aes_alg);
}

static void __exit intel_aes_fini(void)
{
	crypto_unregister_alg(&intel_aes_alg);
}

module_init(intel_aes_init);
module_exit(intel_aes_fini);

MODULE_DESCRIPTION("Rijndael (AES) Cipher Algorithm, Intel AES-NI instructions optimized");
MODULE_LICENSE("GPL");
MODULE_ALIAS("aes");
MODULE_ALIAS("aes-intel");
