/*
 * mtdmap_atca7221.c
 *
 * The map file for Boot Flash.
 *
 * Author: Kazuhiro Yamato <kyamato@miraclelinux.com>
 * 
 * This driver is based on atca7221_flash.c by
 *				Force Computers (India) PVT. LTD.
 *				Soumya K.S <soumya.ks@fci.com>
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation; either version 2 of the License, or (at your
 *	option) any later version.
 *
 *
 *	THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *	WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *	MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *	IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 *	INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *	BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 *	OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *	ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 *	TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 *	USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	You should have received a copy of the GNU General Public License along
 *	with this program; if not, write to the Free Software Foundation, Inc.,
 *	675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <linux/pci.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>

#define MODNAME "mtdmap_atca7221"

#define BOOT_FLASH_PHYS_ADDR	0xFFF00000	/* Boot Flash Area */
#define USER_FLASH_PHYS_ADDR	0xFFE00000	/* User Flash Area */
#define FLASH_SIZE		0x100000

#define PCI_VID_INTEL		0x8086
#define PCI_PID_LPCIF		0x25a1

#define BIOS_CNTL		0x4e

#define BIOS_WRITE_ENABLE	0x0001
#define BIOS_LOCK_ENABLE	0x0002

/*
 * Two partitions have been defined on ATCA-7221
 *
 * Partition 1 (0xFFF00000 - 0xFFFFFFFF)	-- 1.00MB
 * Partition 2 (0xFFE00000 - 0xFFEFFFFF)	-- 1.00MB
 */

#define FLASH_PARTITION_ADDR 0x00000000
#define FLASH_PARTITION_SIZE 0x00100000

struct map_info atca7221_boot_map = {
	.name = "Boot flash device on ATCA-7221",
	.size = FLASH_SIZE,
	.phys = BOOT_FLASH_PHYS_ADDR,
	.bankwidth = 1,
};

struct mtd_partition atca7221_boot_parts[] = {
	{
		name		: "BIOS Image",
		offset		: FLASH_PARTITION_ADDR,
		size		: FLASH_PARTITION_SIZE
	},
};

struct map_info atca7221_user_map = {
	.name = "User flash device on ATCA-7221",
	.phys = USER_FLASH_PHYS_ADDR,
	.size = FLASH_SIZE,
	.bankwidth = 1,
};

struct mtd_partition atca7221_user_parts[] = {
	{
		name		: "USER Flash",
		offset		: FLASH_PARTITION_ADDR,
		size		: FLASH_PARTITION_SIZE
	},
};

#define PARTITION_COUNT_BOOT (sizeof(atca7221_boot_parts)/sizeof(struct mtd_partition))
#define PARTITION_COUNT_USER (sizeof(atca7221_user_parts)/sizeof(struct mtd_partition))

static struct mtd_info *boot_mtd;
static struct mtd_info *user_mtd;

#define	 T_BLOCK_MINUS_BOOT			 0xFFB00002
#define	 T_BLOCK_MINUS_USER			 0xFFA00002

static void __init fwh_init(unsigned int addr)
{
	volatile unsigned char *fwh_tblocks;
	unsigned int count = 0;

	/*
	 * set sectors to be in full access lock down.	This means that the
	 * value of the lock bit cannot be changed, i.e. the chip is always
	 * configured to allow read and write access.
	 */
	for (count = 0; count <= 0xF0000; count += 0x10000) {
		fwh_tblocks = (unsigned char *)ioremap_nocache((addr+count), 4);
		*fwh_tblocks = 0x02;
		iounmap((unsigned char *)fwh_tblocks);
	}
}

static int __init lpcif_init(void)
{
	struct pci_dev *pdev;
	u16 word;

	/*
	 * Search LPCIF
	 */
	pdev = pci_get_device(PCI_VID_INTEL, PCI_PID_LPCIF, NULL);
	if (!pdev){
		printk(KERN_ERR "%s: Cannot find LPCIF(VID:%04x, PID:%04x)\n",
				MODNAME, PCI_VID_INTEL, PCI_PID_LPCIF);
		return -1;
	}

	/*
	 * Read BIOS_CNTL register
	 */
	pci_read_config_word(pdev, BIOS_CNTL, &word);

	/*
	 * Check current bit and set BIOS_WRITE_ENABLE
	 */
	if (!(word & BIOS_WRITE_ENABLE)) {
		if (word & BIOS_LOCK_ENABLE) {
			printk(KERN_ERR "%s: BIOS_LOCK_ENABLE is asserted "
			                "with BIOS_WRITE_DISALBE.\n", MODNAME);
			return -1;
		}
		else {
			word |= BIOS_WRITE_ENABLE;
			pci_write_config_word(pdev, BIOS_CNTL, word);
		}
	}

	return 0;
}

static int __init
init_each_map(unsigned int fwh_addr, char *devname, struct map_info *map,
	       unsigned int phys_addr, struct mtd_info **pmtd,
	       struct mtd_partition parts[], int part_count)
{
	struct mtd_info *mtd;
	fwh_init(fwh_addr);

	printk(KERN_NOTICE "%s on ATCA-7221: %x at %x\n",
			   devname, FLASH_SIZE, phys_addr);

	map->virt = ioremap(phys_addr, FLASH_SIZE);
	if (!map->virt) {
		 printk(KERN_ERR "Failed to ioremap\n");
		 return -EIO;
	}

	*pmtd = do_map_probe("jedec_probe", map);
	mtd = *pmtd;
	if (mtd) {
		(mtd)->owner = THIS_MODULE;

		/*
		 * since the chip cannot be locked make sure the functions are
		 * not configured so that -EOPNOTSUPP is returned when user
		 * attempts to lock an area
		 */
		mtd->lock = NULL;
		mtd->unlock = NULL;
		add_mtd_partitions(mtd, parts, part_count);
		printk(KERN_NOTICE "%s on ATCA-7221 initialized\n", devname);
		return 0;
	}

	iounmap((void *)map->virt);
	printk(KERN_ERR "Failed to install: %s\n", devname);

	return -ENXIO;
}

static int __init init_atca7221(void)
{
	int ret1, ret2;
	if (lpcif_init() < 0)
		return -ENXIO;

	/* BOOT FLASH */
	ret1 = init_each_map(T_BLOCK_MINUS_BOOT, "Boot flash device",
			     &atca7221_boot_map, BOOT_FLASH_PHYS_ADDR,
			     &boot_mtd, atca7221_boot_parts,
			     PARTITION_COUNT_BOOT);

	ret2 = init_each_map(T_BLOCK_MINUS_USER, "User flash device",
			     &atca7221_user_map, USER_FLASH_PHYS_ADDR,
			     &user_mtd, atca7221_user_parts,
			     PARTITION_COUNT_USER);

	return (ret1==0||ret2==0) ? 1 : 0;
}

static void __exit
cleanup_atca7221_each(struct mtd_info *mtd, struct map_info *map)
{
	if (mtd) {
		del_mtd_partitions(mtd);
		map_destroy(mtd);
	}
	if (map->virt) {
		iounmap((void *)map->virt);
		map->virt = 0;
	}
}

static void __exit cleanup_atca7221(void)
{
	cleanup_atca7221_each(boot_mtd, &atca7221_boot_map);
	cleanup_atca7221_each(user_mtd, &atca7221_user_map);
}

module_init(init_atca7221);
module_exit(cleanup_atca7221);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("K. Yamato. <kyamato@miraclelinux.com>");
