/*
 *	RAID Controller driver for Kernel 2.6
 *
 *	(C) Copyright 2001-2009   TOSHIBA SOLUTIONS CORPORATION
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation; either version 2, or (at your option) any
 *	later version.
 *
 */

#define AFC_VERSION "2.06-TOS1(2009-04-24)"

#include	<linux/module.h>
#include	<linux/moduleparam.h>

#include	<linux/types.h>
#include	<linux/string.h>
#include	<linux/kernel.h>
#include	<linux/spinlock.h>
#include	<linux/errno.h>
#include	<linux/ioport.h>
#include	<linux/delay.h>
#include	<linux/timer.h>
#include	<linux/sched.h>
#include	<linux/pci.h>
#include	<linux/pci_ids.h>
#include	<linux/proc_fs.h>
#include	<linux/stat.h>
#include	<linux/slab.h>
#include	<linux/interrupt.h>
#include	<linux/init.h>
#include	<linux/mm.h>
#include	<linux/version.h>
#include	<linux/reboot.h>
#include	<linux/miscdevice.h>

#include	<asm/io.h>
#include	<asm/uaccess.h>
#include	<asm/irq.h>
#include	<asm/byteorder.h>
#include	<asm/processor.h>
#include	<asm/types.h>
#include	<asm/system.h>

#define AFC_PTR_LOW(x) (u32)((unsigned long)(x) & 0x0FFFFFFFF)
#define LinuxVersionCode(v, p, s) (((v) << 16) + ((p) << 8) + (s))

#include	<linux/completion.h>

#include <scsi/scsi.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_host.h>
#include <scsi/scsi_tcq.h>

#ifdef PROPRIETARY_ERROR_HANDLER
#include <scsi/scsi_transport.h>
#endif

#ifndef BITS_PER_LONG
#error "BITS_PER_LONG not defined!"
#endif
#if (BITS_PER_LONG == 64) || defined CONFIG_HIMEM
#define AFC_64BIT_PTR	1
#endif

#include "afxxxc.h"

#if LINUX_VERSION_CODE < LinuxVersionCode(2,6,9)
#error "Kernels older than 2.6.9 are not supported"
#endif


/******************************************************************************
 * SCSI Host template.
 *****************************************************************************/
static struct scsi_host_template AFCSCSI = {
	.module				= THIS_MODULE,
	.proc_info			= afc_scsi_proc_info,
	.proc_name			= PROC_NAME_AFC_SCSI,
	.name				= "TOSHIBA AFC SCSI DRIVER",
	.info				= afc_scsi_info,
	.queuecommand			= afc_scsi_queuecommand ,
	.eh_bus_reset_handler		= afc_scsi_reset,
	.eh_host_reset_handler		= afc_scsi_reset,
	.bios_param			= afc_scsi_bios_param,
	.can_queue			= 0xfffff,
	.this_id			= AFC_SCSI_ID,
	.sg_tablesize			= AFC_SCSI_SGLIST,
	.cmd_per_lun			= 1,
	.use_clustering			= DISABLE_CLUSTERING,
};

#ifdef PROPRIETARY_ERROR_HANDLER
extern void afc_scsi_unjam_host(struct Scsi_Host *);
struct scsi_transport_template AFC_TRANSPORT = {
	.eh_strategy_handler = afc_scsi_unjam_host,
};
#endif

#define	HOST_LOCK	hostdata->host->host_lock

static spinlock_t afc_scsi_lock = SPIN_LOCK_UNLOCKED;


/******************************************************************************
 * PCI device table for all supported controllers.
 *****************************************************************************/
static struct pci_device_id afc_pci_tbl[] = {
	{
		.vendor		= PCI_VENDOR_ID_TOSHIBA,
		.device		= ER34XC_DEVICE_ID,
		.subvendor	= PCI_VENDOR_ID_TOSHIBA,
		.subdevice	= ER340C_SUBDEV_ID,
		0, 0, 0
	},
	{
		.vendor		= PCI_VENDOR_ID_TOSHIBA,
		.device		= AF38XC_DEVICE_ID,
		.subvendor	= PCI_VENDOR_ID_TOSHIBA,
		.subdevice	= T380_SUBDEV_ID,
		0, 0, 1
	},
	{
		.vendor		= PCI_VENDOR_ID_INTEL,
		.device		= HRA38X_DEVICE_ID,
		.subvendor	= 0x1054,
		.subdevice	= HRA380_SUBDEV_ID,
		0, 0, 2
	},
	{
		.vendor		= PCI_VENDOR_ID_TOSHIBA,
		.device		= ER34XC_DEVICE_ID,
		.subvendor	= PCI_VENDOR_ID_TOSHIBA,
		.subdevice	= ER341C_SUBDEV_ID,
		0, 0, 3
	},
	{
		.vendor		= PCI_VENDOR_ID_TOSHIBA,
		.device		= AF38XC_DEVICE_ID,
		.subvendor	= PCI_VENDOR_ID_TOSHIBA,
		.subdevice	= T340_SUBDEV_ID,
		0, 0, 4
	},
	{
		.vendor		= PCI_VENDOR_ID_TOSHIBA,
		.device		= AF38XC_DEVICE_ID,
		.subvendor	= PCI_VENDOR_ID_TOSHIBA,
		.subdevice	= AF380C_SUBDEV_ID,
		0, 0, 5
	},
	{ 0, 0,} ,
};
MODULE_DEVICE_TABLE(pci, afc_pci_tbl);

static struct afc_boards afc_board_tbl[] = {
	{ "ER340C", 2, 0, ER340C_CHANNELS, ER340C_CAN_QUEUE, ER340C_CMD_PER_LUN },
	{ "T380",   3, 0, T380_CHANNELS,   T380_CAN_QUEUE,   T380_CMD_PER_LUN },
	{ "HRA380", 3, 0, HRA380_CHANNELS, HRA380_CAN_QUEUE, HRA380_CMD_PER_LUN },
	{ "ER341C", 3, 0, ER341C_CHANNELS, ER341C_CAN_QUEUE, ER341C_CMD_PER_LUN },
	{ "T340",   3, 0, T340_CHANNELS,   T340_CAN_QUEUE,   T340_CMD_PER_LUN },
	{ "AF380C", 3, 1, AF380C_CHANNELS, AF380C_CAN_QUEUE, AF380C_CMD_PER_LUN },
	{ "      ", 0 }
};

static struct pci_driver afc_pci_driver = {
	.name		= "afxxxc",
	.id_table	= afc_pci_tbl,
	.probe		= afc_probe_one,
	.remove		= __devexit_p(afc_remove_one),
#if LINUX_VERSION_CODE < LinuxVersionCode(2,6,15)
	.driver		= {
		.shutdown	= afc_shutdown
	}
#else
	.shutdown	= afc_shutdown
#endif
};

static struct afc_scsi_host *hosts[MAX_AFC_HOSTS];
static int afc_hosts_max	= MAX_AFC_HOSTS;
static int afc_scsi_hosts	= 0;
static int afc_outgoing_init	= OUTGOING_MAX_INIT;
static int afc_retry_period	= AFC_RETRY_PERIOD;
static int afc_ioctl_timeout	= AFC_IOCTL_TIMEOUT;
static int afc_ioctl_retry_max	= AFC_IOCTL_RETRY_MAX;

static struct timer_list	afc_retry_timer;
static int afc_retry_cnt = 0;

static struct timer_list	afc_stat_timer;

static struct miscdevice	*afc_miscdev = NULL;


static int afc_scsi_proc_info(struct Scsi_Host *host, char *buffer,
	char **start, off_t offset, int length, int inout)
{
	struct afc_scsi_host *hostdata;
	int	i, len;
	off_t	begin = 0;
	int		hostno;

#ifdef AFC_DEBUG_PROC
	off_t	pos = 0;
	unsigned long	*addr;
#endif

	len = 0;

#define PROCMSG(format_args...) len += sprintf(buffer + len, format_args)

	hostno = host->host_no;
	i = host->host_no;
	hostdata = (struct afc_scsi_host *)host->hostdata;

	PROCMSG("    afc card no : %d\n", i);
	PROCMSG("        host no : %d\n", hostno);
	PROCMSG("   pci bus/slot : %d/%d \n", hostdata->pci.dev->bus->number,
		PCI_SLOT(hostdata->pci.dev->devfn));

	PROCMSG("      card type : %s\n",
		afc_board_tbl[hostdata->devnum].name);

	PROCMSG("    max channel : %d\n", hostdata->host->max_channel);
	PROCMSG("can queue depth : %d\n", hostdata->host->can_queue);
	PROCMSG("    cmd per lun : %d\n", hostdata->host->cmd_per_lun);
	PROCMSG(" driver version : %s\n", AFC_VERSION);
	PROCMSG("-\n");

	PROCMSG("the number of cards : %d\n", afc_scsi_hosts);
	if (afc_miscdev)
		PROCMSG(" ioctl device minor : %d\n", afc_miscdev->minor);
	PROCMSG("-\n");

	PROCMSG("      afc_hosts_max : %d\n", afc_hosts_max);
	PROCMSG("      afc_retry_cnt : %d\n", afc_retry_cnt);
	PROCMSG("  afc_outgoing_init : %d\n", afc_outgoing_init);
	PROCMSG("   afc_retry_period : %d\n", afc_retry_period);
	PROCMSG("  afc_ioctl_timeout : %d\n", afc_ioctl_timeout);
	PROCMSG("afc_ioctl_retry_max : %d\n", afc_ioctl_retry_max);
	PROCMSG("-\n");

	PROCMSG("   pending flag : %08x\n", hostdata->pending_flag);
	PROCMSG("     reset flag : %08x\n", hostdata->reset_flag);
	PROCMSG("   post messags : %lu\n", hostdata->post_msgs);
	PROCMSG("  reply messags : %lu\n", hostdata->reply_msgs);
	PROCMSG("    queue depth : %lu\n", hostdata->queue_depth);
	PROCMSG("max queue depth : %lu\n", hostdata->max_queue_depth);
	PROCMSG("  scsi commands : %lu\n", hostdata->scsi_cmds);
	PROCMSG(" queue full sts : %lu\n", hostdata->queue_full_sts);
	PROCMSG("       busy sts : %lu\n", hostdata->busy_sts);
	PROCMSG("      halt code : %08x\n", hostdata->haltcode);

	PROCMSG("-\n");
	PROCMSG("Statistics (lun no: rio/s rsize(KB)/s wio/s wsize(KB)/s)\n");
	for (i = 0; i < MAX_TARGETS; i++) {
		PROCMSG("%2d: %lu %lu %lu %lu\n", i,
		 hostdata->persec_stat[i].r_io_count,
		 hostdata->persec_stat[i].r_io_size,
		 hostdata->persec_stat[i].w_io_count,
		 hostdata->persec_stat[i].w_io_size);
	}

#ifdef AFC_DEBUG_PROC
	PROCMSG("-\n");
	PROCMSG("*Target Info\n");
	for (i = 0; i < MAX_TARGETS; i++) {
		struct afc_target_info *tinfo;
		tinfo = &hostdata->target_info[i];
		PROCMSG("target[%02d] : %d <> %d\n", i,
			tinfo->outgoing_max, tinfo->outgoing_cnt);
	}
	PROCMSG("-\n");
	PROCMSG("*Reply Buffer Info\n");
	addr = (unsigned long *)hostdata->page_frame;
	for (i = 0; i < AFC_REPLY_SIZE * AFC_REPLY_FRAMES / 4; i++) {
		if (*addr) {
			pos = begin + len;
			if (pos < offset) {
				len = 0;
				begin = pos;
			}
			if (pos > offset + length)
				goto out;
			PROCMSG("%p : %08x\n", addr, *addr);
		}
		++addr;
	}
out:

#endif

#undef PROCMSG

	*start = buffer + (offset - begin);	/* Start of wanted data */
	len -= (offset - begin);		/* Start slop */

	if (len > length)
		len = length;			/* Ending slop */

	return len;
}

/******************************************************************************
 *	Functions to handle linked lists of afc scsi command
 *****************************************************************************/
static inline struct afc_scmd *__get_afc_scmd(struct list_head *head)
{
	struct list_head	*lh;

	if (list_empty(head))
		return NULL;

	lh = head->prev;
	list_del(lh);

	return (list_entry(lh, struct afc_scmd, list));
}


static inline struct afc_scmd *get_afc_scmd(struct list_head *head)
{
	struct afc_scmd	*scmd;
	unsigned long	flags;

	spin_lock_irqsave(&afc_scsi_lock, flags);
	scmd = __get_afc_scmd(head);
	spin_unlock_irqrestore(&afc_scsi_lock, flags);

	return (scmd);
}


static inline void __add_afc_scmd(struct afc_scmd  *scmd,
	struct list_head *head)
{
	list_add(&scmd->list, head);

	return;
}


static inline void add_afc_scmd(struct afc_scmd  *scmd,	struct list_head *head)
{
	unsigned long	flags;

	spin_lock_irqsave(&afc_scsi_lock, flags);
	__add_afc_scmd(scmd, head);
	spin_unlock_irqrestore(&afc_scsi_lock, flags);

	return;
}


static inline void __afc_scsi_post_message(struct afc_scsi_host *hostdata,
	struct afc_scmd *scmd, u32 m, unsigned char channel)
{
	__add_afc_scmd(scmd, &hostdata->exec_list[channel]);
	afc_post_message(hostdata, m);

	return;
}


static inline void afc_scsi_post_message(struct afc_scsi_host *hostdata,
	struct afc_scmd  *scmd, u32 m, unsigned char channel)
{
	unsigned long	flags;

	spin_lock_irqsave(&afc_scsi_lock, flags);
	__afc_scsi_post_message(hostdata, scmd, m, channel);
	spin_unlock_irqrestore(&afc_scsi_lock, flags);

	return;
}

/******************************************************************************
 *	AFC scsi reply
 *****************************************************************************/
static void __set_retry_info(struct scsi_cmnd *, struct afc_scsi_host *,
	struct afc_scmd *);
static void set_retry_info(struct scsi_cmnd *, struct afc_scsi_host *,
	struct afc_scmd *);

static void __free_afc_scmd(struct afc_scsi_host *hostdata,
	struct afc_scmd *scmd)
{
	/* statistics */
	if (hostdata->queue_depth)
		hostdata->queue_depth--;
	hostdata->reply_msgs++;

	list_del(&scmd->list);

	/* free scmd */
	scmd->state = SCMD_IDLE;
	scmd->cmd = NULL;
	__add_afc_scmd(scmd, &hostdata->free_list);

	return;
}


static void free_afc_scmd(struct afc_scsi_host *hostdata, struct afc_scmd *scmd)
{
	unsigned long	flags;

	spin_lock_irqsave(&afc_scsi_lock, flags);
	__free_afc_scmd(hostdata, scmd);
	spin_unlock_irqrestore(&afc_scsi_lock, flags);

	return;
}


static void flush_pending_request(struct afc_scsi_host *hostdata,
	struct afc_scmd *reset_scmd, unsigned char channel)
{
	struct afc_scmd		*scmd;
	struct afc_target_info	*tinfo;
	struct scsi_cmnd	*SCpnt;
	unsigned long		flags;
	u8	state;

	do {
		spin_lock_irqsave(&afc_scsi_lock, flags);
		scmd = __get_afc_scmd(&hostdata->exec_list[channel]);
		if (scmd == NULL || scmd == reset_scmd)
			break;
		SCpnt = scmd->cmd;
		state = scmd->state;

		/* statistics */
		hostdata->queue_depth--;

		/* free scmd */
		scmd->cmd = NULL;
		scmd->state = SCMD_IDLE;
		__add_afc_scmd(scmd, &hostdata->free_list);

		spin_unlock_irqrestore(&afc_scsi_lock, flags);

		/* complete scsi command */
		if (SCpnt) {
			spin_lock_irqsave(HOST_LOCK, flags);
			SCpnt->result = DID_RESET << 16;
			SCpnt->scsi_done(SCpnt);
			spin_unlock_irqrestore(HOST_LOCK, flags);

			if (state == SCMD_PENDING) {
				spin_lock_irqsave(&afc_scsi_lock, flags);
				tinfo=&hostdata->target_info[SCpnt->device->id];
				if (tinfo->outgoing_cnt > 0)
					tinfo->outgoing_cnt--;
				afc_retry_cnt--;
				spin_unlock_irqrestore(&afc_scsi_lock, flags);
			}
		}
	} while (1);

	/* clear reset flag */
	hostdata->reset_flag &= ~(0x1<<channel);

	/* statistics */
	hostdata->reply_msgs++;

	if (scmd) {
		/* free scmd */
		reset_scmd->state = SCMD_IDLE;
		reset_scmd->cmd = NULL;
		__add_afc_scmd(reset_scmd, &hostdata->free_list);
		spin_unlock_irqrestore(&afc_scsi_lock, flags);
	} else {
		spin_unlock_irqrestore(&afc_scsi_lock, flags);
		printk(KERN_ERR
			"%s: detected unlinked reset_scmd !!\n",hostdata->name);
	}

	return;
}


static void calc_lun_stat(struct afc_scsi_host *hostdata,
	struct scsi_cmnd *SCpnt)
{
	struct afc_lun_stat	*c_stat;
	unsigned long		flags;

	c_stat = &hostdata->current_stat[SCpnt->device->id];

	if (SCpnt->cmnd[0] == READ_6 || SCpnt->cmnd[0] == READ_10
		|| SCpnt->cmnd[0] == READ_12 || SCpnt->cmnd[0] == READ_16) {
		spin_lock_irqsave(&afc_scsi_lock, flags);
		c_stat->r_io_count++;
		c_stat->r_io_size += SCpnt->request_bufflen;
		spin_unlock_irqrestore(&afc_scsi_lock, flags);
	} else if (SCpnt->cmnd[0] == WRITE_6 || SCpnt->cmnd[0] == WRITE_10
		|| SCpnt->cmnd[0] == WRITE_12 || SCpnt->cmnd[0] == WRITE_16) {
		spin_lock_irqsave(&afc_scsi_lock, flags);
		c_stat->w_io_count++;
		c_stat->w_io_size += SCpnt->request_bufflen;
		spin_unlock_irqrestore(&afc_scsi_lock, flags);
	}

	return;
}


static void free_afc_reset_scmd(struct afc_scsi_host *hostdata,
	struct afc_scmd *scmd, unsigned char channel)
{
	unsigned long	flags;

	spin_lock_irqsave(&afc_scsi_lock, flags);

	list_del(&scmd->list);

	/* clear reset flag */
	hostdata->reset_flag &= ~(0x1<<channel);

	/* statistics */
	hostdata->reply_msgs++;

	/* free scmd */
	scmd->state = SCMD_IDLE;
	scmd->cmd = NULL;
	__add_afc_scmd(scmd, &hostdata->free_list);

	spin_unlock_irqrestore(&afc_scsi_lock, flags);

	return;
}


static void afc_scsi_reply(struct afc_scsi_host *hostdata,
	struct scsi_reply_message *msg)
{
	struct afc_scmd		*scmd;
	struct scsi_cmnd	*SCpnt;
	unsigned long		flags;

	/* get valid scmd */
	do {
		if (msg->key != 0xff)
			break;
		barrier();
	} while (1);

#if (BITS_PER_LONG == 32)
	scmd = (struct afc_scmd *)AFC_PTR_LOW(msg->scmd_ptr);
#else
	scmd = (struct afc_scmd *)msg->scmd_ptr;
#endif

	if (scmd == NULL) {
		printk(KERN_ERR  "%s: detected null scmd on reply\n", hostdata->name);
		return;
	}

	/* get valid SCpnt */
	SCpnt = scmd->cmd;
	if (SCpnt == NULL) {
		printk(KERN_ERR  "%s: detected null SCpnt on reply\n", hostdata->name);
		free_afc_scmd(hostdata, scmd);
		return;
	}

	if (SCpnt->use_sg)
		pci_unmap_sg(hostdata->pci.dev,
			(struct scatterlist *)SCpnt->request_buffer,
			SCpnt->use_sg, SCpnt->sc_data_direction);
	else if (SCpnt->request_bufflen)
		pci_unmap_page(hostdata->pci.dev, SCpnt->SCp.dma_handle,
			SCpnt->request_bufflen, SCpnt->sc_data_direction);

	/* set result */
	if (scmd->message_type == MSG_CDB_EXECUTE) {
		/* set result */
		SCpnt->result = (DID_OK << 16) | msg->status;

		switch (msg->status) {
		case STS_CHECK_CONDITION:
			/* pseudo auto sense */
			if ((SCpnt->cmnd[0] == TEST_UNIT_READY ||
			 SCpnt->cmnd[0] == START_STOP)) {
				memset(SCpnt->sense_buffer,
				 0, sizeof(SCpnt->sense_buffer));
				SCpnt->sense_buffer[0] = 0x70;
				SCpnt->sense_buffer[2] = msg->key;
				SCpnt->sense_buffer[7] = 10;
				SCpnt->sense_buffer[12]= msg->code; /* ASC */
				SCpnt->sense_buffer[13]= msg->qualifier;/*ASCQ*/
				SCpnt->result |= (DRIVER_SENSE << 24);
			}
			break;
		case STS_GOOD:
			if (SCpnt->device->channel == AFC_LOGICAL_CHANNEL)
				if (SCpnt->cmnd[0] != INQUIRY)
					calc_lun_stat(hostdata, SCpnt);
			break;
		case STS_QUEUE_FULL:
			if (SCpnt->device->channel == AFC_LOGICAL_CHANNEL) {
				hostdata->queue_full_sts++;
				scmd->state = SCMD_PENDING;
				set_retry_info(SCpnt, hostdata, scmd);
				return;
			}
			break;
		case STS_BUSY:
			if (SCpnt->device->channel == AFC_LOGICAL_CHANNEL) {
				hostdata->busy_sts++;
				scmd->state = SCMD_PENDING;
				set_retry_info(SCpnt, hostdata, scmd);
				return;
			}
			break;
		case STS_INTERMEDIATE_GOOD:
		case STS_RESERVATION_CONFLICT:
			break;
		case STS_INTERMEDIATE_C_GOOD:
			if (msg->key != 0) {
				printk(KERN_ERR
					"%s: DID_ERROR (k=%x c=%x q=%x)\n",
					hostdata->name,
					msg->key, msg->code, msg->qualifier);
				SCpnt->result = DID_ERROR << 16;
			}
			break;
		case STS_INVALID_HOST_ADAPTER:
		case STS_SELECTION_TIMEOUT:
			SCpnt->result = DID_TIME_OUT << 16;
			break;
		case STS_DATA_OVERRUN_UNDERRUN:
		case STS_UNEXPECTED_BUS_FREE:
		case STS_MESSAGE_REJECT:
			printk(KERN_ERR "%s: DID_ERROR (k=%x c=%x q=%x)\n",
				hostdata->name, msg->key, msg->code, msg->qualifier);
			SCpnt->result = DID_ERROR << 16;
			break;
		case STS_REQUEST_ABORTED:
			if (msg->qualifier == SENSEQUAL_RESET)
				SCpnt->result = DID_RESET << 16;
			else
				SCpnt->result = DID_ABORT << 16;
			break;
		default:
			printk(KERN_ERR "%s: unexpected status(%0x)\n",
				hostdata->name, msg->status);
			SCpnt->result = DID_ERROR << 16;
		}

		spin_lock_irqsave(&afc_scsi_lock, flags);
		if (SCpnt->device->channel == AFC_LOGICAL_CHANNEL) {
			/* evaluate target_info */
			if (hostdata->pending_flag & (1 << SCpnt->device->id)) {
				struct afc_target_info *tinfo;
				tinfo=&hostdata->target_info[SCpnt->device->id];
				tinfo->outgoing_max++;
				tinfo->outgoing_cnt--;
				if (tinfo->outgoing_cnt < 0)
					tinfo->outgoing_cnt = 0;
			}
		}

		/* free scmd */
		__free_afc_scmd(hostdata, scmd);
		spin_unlock_irqrestore(&afc_scsi_lock, flags);

		/* complete scsi command */
		spin_lock_irqsave(HOST_LOCK, flags);
		SCpnt->scsi_done(SCpnt);
		spin_unlock_irqrestore(HOST_LOCK, flags);

	} else if (scmd->message_type == MSG_OUTBOUND_MSG_FLUSH) {
		SCpnt->scsi_done(SCpnt);
	} else if (scmd->message_type == MSG_BUS_RESET) {
		/* scsi bus reset */
		if (msg->status == STS_MSG_COMPLETE) {
			flush_pending_request(
				hostdata, scmd, SCpnt->device->channel);
		} else {
			printk(KERN_ERR
				"%s: cannot complete bus reset\n", hostdata->name);
			free_afc_reset_scmd(
				hostdata, scmd, SCpnt->device->channel);
		}
	} else {
		printk(KERN_ERR
			"%s: detected invalid message status\n",hostdata->name);
		/* free scmd */
		free_afc_scmd(hostdata, scmd);
	}

	return;
}


/* AFC interrupt handler */
static irqreturn_t afc_pci_interrupt(int irq, void *dev_id, struct pt_regs *r)
{
	struct afc_scsi_host *hostdata = dev_id;
	struct scsi_reply_message *m;
	unsigned long mbus, madt;
	irqreturn_t handled;

	if ((mbus = AFC_REPLY_READ32(hostdata)) == 0xFFFFFFFF)
		mbus = AFC_REPLY_READ32(hostdata);

	handled = IRQ_NONE;
	while (mbus != 0xFFFFFFFF) {
		madt = (mbus & 0x03FFFFFF) -
			(hostdata->page_frame_bus & 0x03FFFFFF);
		madt += (unsigned long)hostdata->page_frame;
		m = (struct scsi_reply_message *)madt;

#ifdef AFC_DEBUG_IRQ
		printk("%s: reply_addr=%08x\n", hostdata->name, mbus);
		printk("%s: msg[0]=%08x\n", hostdata->name, *(u32*)m);
		printk("%s: msg[1]=%08x\n", hostdata->name, *((u32*)m+1));
		printk("%s: msg[2]=%08x\n", hostdata->name, *((u32*)m+2));
		printk("%s: msg[3]=%08x\n", hostdata->name, *((u32*)m+3));
#endif
		afc_scsi_reply(hostdata, m);

		/* clear and flush msg */
		memset(m, (int)0xFFFFFFFF, sizeof(struct scsi_reply_message));
		afc_flush_reply(hostdata, mbus);
		mb();

		if ((mbus = AFC_REPLY_READ32(hostdata)) == 0xFFFFFFFF)
			mbus = AFC_REPLY_READ32(hostdata);
		handled = IRQ_HANDLED;
	}

	return (handled);
}


/* writeback cache on/off and shutdown */
static void set_modesel_cmnd(struct scsi_cmnd *SCpnt, struct Scsi_Host *host,
	struct afc_parameter *pbuf)
{
	SCpnt->device->host	 = host;
	SCpnt->device->channel	 = 0;
	SCpnt->device->id	 = 0;
	SCpnt->device->lun	 = 0;

	SCpnt->cmnd[0]		 = MODE_SELECT;
	SCpnt->cmnd[1]		 = 0x11; /* LUN=0, PF=1, SP=1 */
	SCpnt->cmnd[2]		 = 0;
	SCpnt->cmnd[3]		 = 0;
	SCpnt->cmnd[4]		 = sizeof(struct afc_parameter);
	SCpnt->cmnd[5]		 = 0;

	SCpnt->use_sg		 = 0;
	SCpnt->sc_data_direction = DMA_TO_DEVICE;
	SCpnt->request_buffer	 = pbuf;
	SCpnt->request_bufflen	 = sizeof(struct afc_parameter);

	return;
}


static void set_cache_mode(struct afc_parameter *pbuf, u_char mode)
{
	pbuf->head.block_desc_length	= 0x08;
	pbuf->head.block_length_med	= 0x02;
	pbuf->para.page3c.PageCode	= 0x3C;
	pbuf->para.page3c.ParaLength	= sizeof(struct s_Page3Cpara) - 2;
	pbuf->para.page3c.Mode		= mode;

	return;
}


static void afc_cache_mode(struct afc_scsi_host *hostdata, u_char mode)
{
	memset(&hostdata->control_device, 0, sizeof(struct scsi_device));
	memset(&hostdata->control_cmnd,   0, sizeof(struct scsi_cmnd));
	memset(&hostdata->parameter_buf,  0, sizeof(struct afc_parameter));

	hostdata->control_cmnd.device = &hostdata->control_device;
	set_cache_mode(&hostdata->parameter_buf, mode);
	set_modesel_cmnd(&hostdata->control_cmnd, hostdata->host,
		&hostdata->parameter_buf);

	afc_scsi_command(&hostdata->control_cmnd);
	if (host_byte(hostdata->control_cmnd.result) != DID_OK) {
		printk(KERN_ERR "%s: unable to set cache mode(%0x)\n",
			hostdata->name, mode);
	}

	return;
}


/* AFC IOCTL handler */
static void invalidate_ioctl_scmd(struct afc_scsi_host *hostdata,
	struct scsi_cmnd *SCpnt)
{
	struct list_head	*lh;
	struct afc_scmd		*scmd;
	unsigned long		flags;

	spin_lock_irqsave(&afc_scsi_lock, flags);
	for (lh = hostdata->exec_list[SCpnt->device->channel].prev;
		lh != &hostdata->exec_list[SCpnt->device->channel]; lh = lh->prev) {
		scmd = list_entry(lh, struct afc_scmd, list);
		if (scmd->cmd == SCpnt) {
			scmd->cmd = NULL;
		}
	}
	spin_unlock_irqrestore(&afc_scsi_lock, flags);

	return;
}


static void ioctl_halt_done(struct scsi_cmnd * SCpnt)
{
	printk(KERN_ERR "afxxxc: ioctl_halt_done called\n");

	return;
}


static void ioctl_timeout(unsigned long data)
{
	struct scsi_cmnd	*SCpnt = (struct scsi_cmnd *)data;
	struct Scsi_Host	*host  =
					(struct Scsi_Host *)SCpnt->device->host;
	struct afc_scsi_host	*hostdata =
					(struct afc_scsi_host *)host->hostdata;
	struct timer_list	*timeout_timer;
	unsigned long		flags;

	SCpnt->timeout_per_command = 1;
	timeout_timer = (struct timer_list *)SCpnt->request->buffer;

	spin_lock_irqsave(HOST_LOCK, flags);
	if (SCpnt->result)
		goto clear_and_return;

	if (hostdata->haltcode) {
		invalidate_ioctl_scmd(hostdata, SCpnt);
		SCpnt->result = hostdata->haltcode;
		SCpnt->scsi_done(SCpnt);
		SCpnt->scsi_done = ioctl_halt_done;
		goto clear_and_return;
	}

	if (SCpnt->retries++ == 0) {
		printk(KERN_INFO
			"%s: ioctl timeout detected. (%02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x)\n",
			hostdata->name,
			SCpnt->cmnd[0], SCpnt->cmnd[1], SCpnt->cmnd[2],
			SCpnt->cmnd[3], SCpnt->cmnd[4], SCpnt->cmnd[5],
			SCpnt->cmnd[6], SCpnt->cmnd[7], SCpnt->cmnd[8],
			SCpnt->cmnd[9], SCpnt->cmnd[10],SCpnt->cmnd[11]);
		afc_scsi_reset(SCpnt);
		mod_timer(timeout_timer, jiffies + afc_ioctl_timeout);
	} else if (SCpnt->retries < afc_ioctl_retry_max) {
		switch (afc_board_tbl[hostdata->devnum].generation) {
		case 3:
			pci_read_config_dword(hostdata->pci.dev,
				IOP348_HALTCODE_REG, &hostdata->haltcode);
			break;
		case 2:
			pci_read_config_dword(hostdata->pci.dev,
				IOP33X_HALTCODE_REG, &hostdata->haltcode);
			break;
		default:
			break;
		}
		if (hostdata->haltcode) {
			printk(KERN_ERR "afxxxc: %s is halted (%08x)\n",
				hostdata->name, hostdata->haltcode);
			mod_timer(timeout_timer, jiffies + HZ * 2);
		} else {
			mod_timer(timeout_timer, jiffies+ afc_ioctl_timeout*2);
		}
	}

clear_and_return:
	spin_unlock_irqrestore(HOST_LOCK, flags);
	SCpnt->timeout_per_command = 0;

	return;
}


static void ioctl_done(struct scsi_cmnd * SCpnt)
{
	if (SCpnt->request->buffer)
		del_timer_sync((struct timer_list *)SCpnt->request->buffer);
	complete(SCpnt->request->waiting);

	return;
}


static void ioctl_do_command(struct scsi_cmnd *SCpnt)
{
	struct Scsi_Host *host = SCpnt->device->host;
	struct afc_scsi_host *hostdata = (struct afc_scsi_host *)host->hostdata;
	struct timer_list timeout_timer;
	unsigned long flags;
	DECLARE_COMPLETION(wait);

	init_timer(&timeout_timer);
	timeout_timer.data	= (unsigned long)SCpnt;
	timeout_timer.function	= ioctl_timeout;

	SCpnt->request->waiting = &wait;
	SCpnt->request->buffer  = (char *)&timeout_timer;

	spin_lock_irqsave(HOST_LOCK, flags);
	timeout_timer.expires = jiffies + afc_ioctl_timeout;
	add_timer(&timeout_timer);

	afc_scsi_queuecommand(SCpnt, ioctl_done);

	spin_unlock_irqrestore(HOST_LOCK, flags);

	wait_for_completion(&wait);

	while (SCpnt->timeout_per_command) {
		barrier();
	}

	return;
}


static unsigned char request_sense_cmnd[6] = {REQUEST_SENSE, 0,0,0, 16, 0};

static void ioctl_do_request_sense(struct scsi_cmnd *SCpnt)
{
	struct Scsi_Host *host = SCpnt->device->host;
	struct afc_scsi_host *hostdata = (struct afc_scsi_host *)host->hostdata;
	int	result;

	memcpy((void *)SCpnt->cmnd, (void *)request_sense_cmnd,
		sizeof(request_sense_cmnd));

	SCpnt->request_buffer = &SCpnt->sense_buffer;
	SCpnt->request_bufflen = sizeof(SCpnt->sense_buffer);
	SCpnt->use_sg = 0;

	SCpnt->cmd_len = COMMAND_SIZE(SCpnt->cmnd[0]);
	SCpnt->result = 0;

	if (hostdata->haltcode == 0) {
		ioctl_do_command(SCpnt);

		result  = 0x00FF0000;
		result |= ((int)SCpnt->sense_buffer[2])  << 24;
		result |= ((int)SCpnt->sense_buffer[12]) <<  8;
		result |= ((int)SCpnt->sense_buffer[13]) <<  0;
		SCpnt->result = result;
	} else {
		SCpnt->result = hostdata->haltcode;
	}

	return;
}


static int afc_scsi_ioctl_send_command(int host_id, unsigned long data)
{
	struct afc_scsi_host	*hostdata;
	struct scsi_device	*SDpnt;
	struct scsi_cmnd	*SCpnt;
	struct request		*req;
	struct afc_ioctl_command	*sic;
	void __user 		*arg = (void __user *)data;
	char			*buf;
	unsigned char		cmd[MAX_COMMAND_SIZE];
	char			*cmd_in;
	unsigned char		opcode;
	unsigned int		inlen, outlen, cmdlen;
	int			buf_needed, result;
	unsigned char		channel, target;

	sic = (struct afc_ioctl_command *)arg;
	if (!sic)
		return -EINVAL;

	if (host_id >= afc_scsi_hosts)
		return -EINVAL;
	hostdata = hosts[host_id];

	/* Verify afc_ioctl_command */
	result = access_ok(VERIFY_READ, sic, sizeof(struct afc_ioctl_command));
	if (!result) return result;

	get_user(inlen,	&sic->inlen);
	get_user(target, &sic->target);
	get_user(channel, &sic->channel);
	get_user(outlen, &sic->outlen);

	/* We do not transfer more than MAX_IOCTL_BUFF */
	if (inlen  > MAX_IOCTL_BUFF)
		return -EINVAL;
	if (outlen > MAX_IOCTL_BUFF)
		return -EINVAL;

	/* Check target and channel number */
	if (target >= MAX_TARGETS)
		return -EINVAL;
	if (channel >= hostdata->host->max_channel)
		return -EINVAL;

	cmd_in = sic->data;
	get_user(opcode, cmd_in);

	buf_needed = (inlen > outlen ? inlen : outlen);
	if (buf_needed) {
		if ((buf = (char *)kmalloc(buf_needed, GFP_KERNEL)) == NULL)
			return -ENOMEM;
		memset(buf, 0, buf_needed);
	} else {
		buf = NULL;
	}

	/* Get the command from the user's address space */
	cmdlen = COMMAND_SIZE(opcode);

	result = access_ok(VERIFY_READ, cmd_in, cmdlen + inlen);
	if (!result) {
		if (buf)
			kfree(buf);
		return result;
	}

	result = copy_from_user((void *)cmd, cmd_in, cmdlen);

	/* Get the data to be sent to the device */
	result = copy_from_user((void *)buf, cmd_in + cmdlen, inlen);

	/* Get Scsi Request Buffer */
	SDpnt = (struct scsi_device *)
		kmalloc(sizeof(struct scsi_device), GFP_KERNEL);
	if (!SDpnt) {
		if (buf)
			kfree(buf);
		return -ENOMEM;
	}

	SCpnt =(struct scsi_cmnd *)kmalloc(sizeof(struct scsi_cmnd),GFP_KERNEL);
	if (!SCpnt) {
		kfree(SDpnt);
		if (buf)
			kfree(buf);
		return -ENOMEM;
	}

	req = (struct request *)kmalloc(sizeof(struct request), GFP_KERNEL);
	if (!req) {
		kfree(SCpnt);
		kfree(SDpnt);
		if (buf)
			kfree(buf);
		return -ENOMEM;
	}

	memset(SDpnt, 0, sizeof(struct scsi_device));
	memset(SCpnt, 0, sizeof(struct scsi_cmnd));
	memset(req, 0, sizeof(struct request));

	/* Set Scsi Cmnd Buffer */
	SCpnt->device		= SDpnt;
	SCpnt->device->host	= hostdata->host;
	SCpnt->device->channel	= channel;
	SCpnt->device->id	= target;
	SCpnt->device->lun	= 0;
	SCpnt->request		= req;

	memcpy(SCpnt->cmnd, cmd, cmdlen);
	SCpnt->cmd_len		= cmdlen;
	SCpnt->use_sg		= 0;
	SCpnt->request_buffer	= buf;
	SCpnt->request_bufflen	= buf_needed;

	if (hostdata->haltcode == 0)
		ioctl_do_command(SCpnt);
	else
		SCpnt->result = hostdata->haltcode;

	if (SCpnt->result) {
		if (status_byte(SCpnt->result) == CHECK_CONDITION)
			ioctl_do_request_sense(SCpnt);
		result = access_ok(VERIFY_WRITE, cmd_in, 4);
		if (!result)
			goto free_and_return;
		result = copy_to_user((void *)cmd_in, &SCpnt->result, 4);
		result = -EIO;
	} else {
		result = access_ok(VERIFY_WRITE, cmd_in, outlen);
		if (!result)
			goto free_and_return;
		result = copy_to_user((void *)cmd_in, buf, outlen);
		result = 0;
	}

free_and_return:
	kfree(req);
	kfree(SCpnt);
	kfree(SDpnt);
	if (buf)
		kfree(buf);

	return result;
}


static int cfg_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	unsigned long arg)
{
	if (_IOC_TYPE(cmd) != AFC_MAGIC_NUMBER) {
		printk(KERN_ERR "afxxxc: ioctl command failed.\n");
		return -ENOTTY;
	}
	if ((_IOC_NR(cmd) < AFC_MINNR) && (_IOC_NR(cmd) > AFC_MAXNR)) {
		printk(KERN_ERR "afxxxc: ioctl command failed.\n");
		return -ENOTTY;
	}

	switch(cmd) {
		case AFC_SCSI_SEND_COMMAND_0:
			return afc_scsi_ioctl_send_command(0, arg);
		case AFC_SCSI_SEND_COMMAND_1:
			return afc_scsi_ioctl_send_command(1, arg);
		case AFC_SCSI_SEND_COMMAND_2:
			return afc_scsi_ioctl_send_command(2, arg);
		case AFC_SCSI_SEND_COMMAND_3:
			return afc_scsi_ioctl_send_command(3, arg);

		default:
			return -EINVAL;
	}
}


static long long cfg_llseek(struct file *file, long long offset, int origin)
{
	return -ESPIPE;
}


static ssize_t cfg_write(struct file *file, const char *buf, size_t count,
	loff_t *ppos)
{
	return 0;
}


static ssize_t cfg_read(struct file *file, char *buf, size_t count, loff_t *ptr)
{
	return 0;
}


static int cfg_open(struct inode *inode, struct file *file)
{
	return 0;
}


static int cfg_release(struct inode *inode, struct file *file)
{
	return 0;
}


static struct file_operations config_fops =
{
	owner:		THIS_MODULE,
	llseek: 	cfg_llseek,
	read:		cfg_read,
	write:		cfg_write,
	ioctl:		cfg_ioctl,
	open:		cfg_open,
	release:	cfg_release,
};

static struct miscdevice miscdev = {
	AFC_MINOR ,
	"afc" ,
	&config_fops
};


static int afc_install_ioctl(void)
{
	if (afc_miscdev)
		return -EINVAL;

	if (misc_register(&miscdev) == -1) {
		printk(KERN_ERR "afxxxc: can't register ioctl device.\n");
		return -EBUSY;
	}

	afc_miscdev = &miscdev;
	return 0;
}


static int afc_remove_ioctl(void)
{
	if (afc_miscdev != &miscdev)
		return -EINVAL;

	misc_deregister(afc_miscdev);
	afc_miscdev = NULL;

	return 0;
}


/* Flush outbound message */
static void make_header(struct scsi_cmnd *, struct afc_scsi_host *,
	struct scsi_post_message *, struct afc_scmd *, u8);


static void flush_outbound_message(struct afc_scsi_host *hostdata)
{
	struct scsi_cmnd			*SCpnt;
	struct scsi_device			*SDpnt;
	struct request				*req;
	struct afc_scmd				*scmd;
	struct scsi_post_message	*msg;
	u32	m;
	DECLARE_COMPLETION(wait);

	/* set SCpnt */
	SDpnt = &hostdata->control_device;
	SCpnt = &hostdata->control_cmnd;
	req   = &hostdata->control_req;
	memset(SDpnt, 0, sizeof(struct scsi_device));
	memset(SCpnt, 0, sizeof(struct scsi_cmnd));
	memset(req, 0, sizeof(struct request));

	SCpnt->device		= SDpnt;
	SCpnt->device->host	= hostdata->host;
	SCpnt->device->channel	= 0;
	SCpnt->device->id	= 0;
	SCpnt->device->lun	= 0;
	SCpnt->request		= req;
	SCpnt->request->waiting = &wait;
	SCpnt->use_sg		= 0;
	SCpnt->scsi_done	= ioctl_done;

	/* get inbound message */
	m = AFC_POST_READ32(hostdata);
	if (m == 0xFFFFFFFF) {
		printk(KERN_ERR "%s: unable to flush outbound messages\n",
			hostdata->name);
		return;
	}
	msg = (struct scsi_post_message *)(hostdata->mem_offset + m);

	/* get free scmd */
	scmd = get_afc_scmd(&hostdata->free_list);
	if (scmd == NULL) {
		printk(KERN_ERR "%s: unable to flush outbound messages due to lack of scmd.\n",
			hostdata->name);
		return;
	}

	/* set scmd */
	scmd->state = SCMD_POSTED;
	scmd->cmd = SCpnt;
	scmd->message_type = MSG_OUTBOUND_MSG_FLUSH;

	/* set header */
	make_header(SCpnt, hostdata, msg, scmd, MSG_OUTBOUND_MSG_FLUSH);

	afc_scsi_post_message(hostdata, scmd, m, 0);

	wait_for_completion(&wait);

	return;
}


static struct scsi_reply_message *allocate_reply_buffer(struct pci_dev *pdev,
	dma_addr_t *mbuf_bus)
{
	struct scsi_reply_message *mbuf, *free_mbuf;
	dma_addr_t	free_bus;
	size_t	mbuf_size;

	mbuf_size = AFC_REPLY_SIZE * AFC_REPLY_FRAMES;
	mbuf = (struct scsi_reply_message *)
		dma_alloc_coherent(&pdev->dev, mbuf_size, mbuf_bus, GFP_ATOMIC);

	if (mbuf == NULL)
		return NULL;

	if ((u64)(*mbuf_bus + mbuf_size) >
		((u64)*mbuf_bus & 0xFFFFFFFFFC000000ull) + 0x4000000ull) {
		free_mbuf = mbuf;
		free_bus = *mbuf_bus;

		mbuf = allocate_reply_buffer(pdev, mbuf_bus);

		dma_free_coherent(&pdev->dev, mbuf_size, (void *)free_mbuf, free_bus);

		return mbuf;
	}

	return mbuf;
}


static int afc_scsi_ioctl_send_command(int, unsigned long);
static void afc_retry_queuecommand(unsigned long);
static void afc_update_lun_stat(unsigned long);


const char *afc_scsi_info(struct Scsi_Host *host)
{
	static char buffer[AFC_SCSI_INFO];
	struct afc_scsi_host *hostdata;

	hostdata = (struct afc_scsi_host *)host->hostdata;

	snprintf(buffer, AFC_SCSI_INFO,
		"%s %s RAID Controller [%s]"
		,hostdata->pci.did == HRA38X_DEVICE_ID ? "HITACHI" : "TOSHIBA"
		,afc_board_tbl[hostdata->devnum].name, AFC_VERSION);

	return (buffer);
}


/* AFC queuecommand */
static int check_logical_block_address(u8 *cmd)
{
	/* Following commands which has logical address fields are not defined
	 *  in <scsi/scsi.h>
	 * REBUILD, REGNERATE, XDREAD, XDWRITE, XDWRITE_16, XPWRITE
	 * PREFETCH_16, SYCHRONIZED_CACHE_16, VERIFY_12, WRITE_VERIFY_16, XDREADWRITE */
	switch (*cmd)  {
	/* SCSI-2 Direct Access */
	case LOCK_UNLOCK_CACHE: case PRE_FETCH:		case READ_6:
	case READ_10:		case READ_CAPACITY:	case READ_LONG:
	case SEARCH_EQUAL:	case SEARCH_HIGH:	case SEARCH_LOW:
	case SEEK_6:		case SEEK_10:		case SET_LIMITS:
	case SYNCHRONIZE_CACHE: case VERIFY:		case WRITE_6:
	case WRITE_10:		case WRITE_VERIFY:	case WRITE_LONG:
	case WRITE_SAME:
	/* SCSI-2 Write Once and Optical ( for AF100C ) */
	case MEDIUM_SCAN:	case SEARCH_EQUAL_12:	case SEARCH_HIGH_12:
	case SEARCH_LOW_12:	case UPDATE_BLOCK:	case WRITE_VERIFY_12:
	/* SPC-2 & SBC */
	case READ_12:		case WRITE_12:
	/* SPC-3 & SBC-2 */
	case READ_16:
	case SERVICE_ACTION_IN: /* (READ_CAPACITY_16, READ_LONG_16, WRITE_LONG_16) */
	case VERIFY_16:		case WRITE_16:
		return 1;
	default:
		return 0;
	}
}


static u32 calc_block_address(u8 *cmd)
{
	u32 block_addr = 0;

	/* calculate block address */
	switch (*cmd >> 5) {
	case 0:
		block_addr |= (*(cmd + 1) & 0x1f) << 16;
		block_addr |= *(cmd + 2) << 8;
		block_addr |= *(cmd + 3);
		break;
	case 1:
	case 2:	/* 10 Byte CDB */
	case 5:	/* 12 Byte CDB */
		block_addr |= *(cmd + 2) << 24;
		block_addr |= *(cmd + 3) << 16;
		block_addr |= *(cmd + 4) << 8;
		block_addr |= *(cmd + 5);
		break;
	case 4:	/* 16 Byte CDB */
		block_addr |= *(cmd + 6) << 24;
		block_addr |= *(cmd + 7) << 16;
		block_addr |= *(cmd + 8) << 8;
		block_addr |= *(cmd + 9);
		break;
	default:
		return 0;
	}
	/* convert logical block address to byte offset */
	/* presumption-1 : afc sector size = 512B */
	/* presumption-2 : afc cache stripe size = 16KB (32 sectors) */
	block_addr = (block_addr & 0x1f) * 512;

	return block_addr;
}


static struct scsi_post_message *get_post_message(
	struct afc_scsi_host *hostdata)
{
	u32	m;

	do {
		mb();
		m = AFC_POST_READ32(hostdata);
	} while (m == 0xFFFFFFFF);

	return (struct scsi_post_message *)(hostdata->mem_offset + m);
}


static void make_header(struct scsi_cmnd *SCpnt, struct afc_scsi_host *hostdata,
	struct scsi_post_message *msg, struct afc_scmd *scmd, u8 message)
{
	u8 bus_no, target_id, lun;

	lun			= SCpnt->device->lun;
	target_id	= SCpnt->device->id;
	bus_no		= SCpnt->device->channel;

	if (target_id > 7 && bus_no == AFC_LOGICAL_CHANNEL)
		target_id --;

	msg->next_ptr	= (u32)NULL;
#ifdef AFC_64BIT_PTR
	msg->scmd_ptr	= (u64)scmd;
#else
	msg->scmd_ptr	= (u64)((u32)scmd);
#endif
	msg->dma_count	= 0;
	msg->message	= message;
	msg->number		= 0;
	msg->length		= 1;

	if (message == MSG_CDB_EXECUTE && SCpnt->device != NULL &&
		SCpnt->device->tagged_supported) {
		if (SCpnt->cmnd[0] != REQUEST_SENSE) {
			SCpnt->tag = scmd->tag_no;
			msg->queue_code = SIMPLE_QUEUE_TAG;
			msg->queue_tag = SCpnt->tag;
		} else {
			msg->queue_tag = 0;
			msg->queue_code = 0;
		}
	} else {
		msg->queue_tag = 0;
		msg->queue_code = 0;
	}

	msg->lun = lun;
	msg->target_id = target_id;
	msg->bus_no = bus_no;

	if (afc_board_tbl[hostdata->devnum].cdb16_support) {
		memcpy(msg->u.cdb16.cdb, SCpnt->cmnd, 16);
		msg->u.cdb16.cdb_size = SCpnt->cmd_len;
		msg->u.cdb16.busy_status = 0;
		msg->u.cdb16.check_condition = 0;
		msg->u.cdb16.ref_count = 0;
	} else {
		memcpy(msg->u.cdb12.cdb, SCpnt->cmnd, 12);
		msg->u.cdb12.cdb_size = SCpnt->cmd_len;
		msg->u.cdb12.busy_status = 0;
		msg->u.cdb12.check_condition = 0;
		msg->u.cdb12.ref_count = 0;
	}

	return;
}


#ifndef min
#define min(a,b) ((a)<(b) ? (a) : (b))
#endif

static struct scsi_post_message *create_sg_list(struct scsi_cmnd *SCpnt,
	struct afc_scsi_host *hostdata, struct scsi_post_message *head,
	u32 divide, u32 align, u32 block_addr)
{
	struct scsi_post_message *tail, *msg;
	struct scatterlist *sg, sglist;
	struct page *page;
	dma_addr_t dma_addr;
	unsigned long	offset;
	u32	len, count, *mptr;
	u8	number, dma_count;
	int	i;
	u16	sg_count;

	if (SCpnt->use_sg) {
		sg = (struct scatterlist *)SCpnt->request_buffer;
		sg_count = pci_map_sg(hostdata->pci.dev, sg, SCpnt->use_sg,
			SCpnt->sc_data_direction);
	} else {
		page = virt_to_page(SCpnt->request_buffer);
		offset = offset_in_page(SCpnt->request_buffer);
		sg_dma_address(&sglist) = pci_map_page(hostdata->pci.dev, page,
			offset, SCpnt->request_bufflen, SCpnt->sc_data_direction);
		sg_dma_len(&sglist)  = SCpnt->request_bufflen;
		SCpnt->SCp.dma_handle = sg_dma_address(&sglist);
		sg = &sglist;
		sg_count = 1;
	}

	tail = head;
	mptr = (u32 *)head + 8;
	number = dma_count = 0;

	for (i = 0; i < sg_count; i++) {
		dma_addr = sg_dma_address(sg);
		len = sg_dma_len(sg);

		do {
			if (divide)
				count = min(len, (align - block_addr));
			else
				count = len;

			/* dma request */
#ifdef AFC_64BIT_PTR
			*mptr = (count << 8) | ((dma_addr >> 32) & 0xff);
			++ mptr;
			*mptr = (u32)(dma_addr & 0xffffffff);
			++ mptr;
#else
			*mptr = count << 8;
			++ mptr;
			*mptr = dma_addr;
			++ mptr;
#endif

			/* update counter */
			if (divide)
				block_addr = (block_addr + count) & (align - 1);
			dma_count ++;
			len -= count;
			dma_addr += count;

			/* append new message if necessary */
			if (dma_count == AFC_MAX_SGLISTS &&
				(len || i + 1 < sg_count)) {
				/* get scsi message buffer */
				msg = get_post_message(hostdata);
				/* set left part of header in tail */
#if (BITS_PER_LONG == 32)
				tail->next_ptr =
					((unsigned long)msg - hostdata->mem_offset);
#else
				tail->next_ptr = AFC_PTR_LOW(
					((unsigned long)msg - hostdata->mem_offset));
#endif
				tail->number = number;
				tail->dma_count = AFC_MAX_SGLISTS;
				tail = msg;

				/* set header in msg */
				memcpy(msg, head, 32);

				/* update variables */
				mptr = (u32 *)msg + 8;
				++ number;
				dma_count = 0;
			}
		} while (len);

		sg ++;
	}

	/* set left part of header in tail */
	tail->next_ptr = (u32)NULL;
	tail->number = number++;
	tail->dma_count = dma_count;

	/* set length field in head */
	head->length = number;

	return (head);
}


static struct scsi_post_message *create_scsi_list(struct scsi_cmnd * SCpnt,
	struct afc_scsi_host *hostdata, struct afc_scmd *scmd)
{
	struct scsi_post_message *msg_head;
	u32	divide, align, block_addr;

	/* initialize */
	if (SCpnt->device->channel == AFC_LOGICAL_CHANNEL &&
		check_logical_block_address(SCpnt->cmnd)) {
		divide = 1;
		block_addr = calc_block_address(SCpnt->cmnd);
		align = 0x4000;
	} else {
		divide = 0;
		align = 0;
		block_addr = 0;
	}

	/* get scsi message buffer */
	msg_head = get_post_message(hostdata);

	/* set header */
	make_header(SCpnt, hostdata, msg_head, scmd, MSG_CDB_EXECUTE);

	/* create scsi message list */
	if (SCpnt->use_sg || SCpnt->request_bufflen) {
		msg_head = create_sg_list(SCpnt,hostdata,msg_head,divide,align,
			block_addr);
	}

	return (msg_head);
}


static void __do_scsi_queuecommand(struct scsi_cmnd * SCpnt,
	struct afc_scsi_host *hostdata, struct afc_scmd *scmd)
{
	struct scsi_post_message *msg_head;
	u8	length;

	/* create inbound messages and send them */
	msg_head = create_scsi_list(SCpnt, hostdata, scmd);
	length = msg_head->length;

	afc_post_message(hostdata,(u32)((unsigned long)msg_head- hostdata->mem_offset));

	/* statistics */
	hostdata->post_msgs += length;

	return;
}

int afc_is_quick_return_command(struct scsi_cmnd *SCpnt,
				struct afc_scsi_host *hostdata)
{
	int	unsupported = 0;
	int	sense_code = 0;
	struct	scatterlist	*sg;
	char	*virtual_address = NULL;

	if (SCpnt->use_sg > 0) {
		sg = (struct scatterlist *)SCpnt->request_buffer;
		if (sg->page) {
			virtual_address = (char *)
				page_address(sg->page) +
					sg->offset;
		}
	}

	switch (SCpnt->cmnd[0]) {
	case MODE_SENSE:
	case MODE_SENSE_10:
		if (((SCpnt->cmnd[2] & 0x3f) == 0x19)
			|| ((SCpnt->cmnd[2] & 0x3f) == 0x1c)) {
			sense_code = 0x2400;
			unsupported = 1;
		}
		break;
	case LOG_SENSE:
		sense_code = 0x2000;
		unsupported = 1;
		break;
	default:
		break;
	}

	if (unsupported == 1) {
		memset(SCpnt->sense_buffer, 0, sizeof(SCpnt->sense_buffer));
		SCpnt->sense_buffer[0] = 0x70;
		SCpnt->sense_buffer[2] = 0x05;	/* Sense Key */
		SCpnt->sense_buffer[7] = 10;	/* Additional Sense Data Length */
		SCpnt->sense_buffer[12] = (sense_code >> 8);
						/* Additional Sense Code */
		SCpnt->sense_buffer[13] = (sense_code & 0xff);
						/* Additional Sense Qualifier */
		SCpnt->request_bufflen = sizeof(SCpnt->sense_buffer);

		if (SCpnt->use_sg > 0)
			memcpy(virtual_address, SCpnt->sense_buffer,
					sizeof(SCpnt->sense_buffer));
		else
			SCpnt->request_buffer = &SCpnt->sense_buffer;

		SCpnt->result = (DRIVER_SENSE << 24) | STS_CHECK_CONDITION;
		SCpnt->scsi_done(SCpnt);
		return 1;
	}

	return 0;
}


int afc_scsi_queuecommand(struct scsi_cmnd * SCpnt,
	void (*done)(struct scsi_cmnd *))
{
	struct Scsi_Host *host = SCpnt->device->host;
	struct afc_scsi_host *hostdata = (struct afc_scsi_host *)host->hostdata;
	struct afc_scmd *scmd;
	unsigned long flags;

/*
	I wonder this patch is suitable. Because patching to scsi_error.c/
	scsi_send_eh_cmnd(), request sense execute with SCpnt->use_sg = 1.
	But receiving process doesn't accept scatter list format of sense
	data. So, scsi_eh_get_sense() is always failed.
	To avoid this, intentionally set to SCpnt->use_sg = 0 and clear
	scatter list in SCpnt->request_buffer.
	2008-10-21
*/
	if (SCpnt->cmnd[0] == REQUEST_SENSE) {
		if (SCpnt->use_sg > 0 && SCpnt->device->host->eh_action != NULL) {
			SCpnt->use_sg = 0;
			SCpnt->request_buffer = &(SCpnt->sense_buffer[0]);
		}
	}

	/* initialize */
	SCpnt->scsi_done = done;

	/* sanity check */
	if (SCpnt->device->id >= MAX_TARGETS) {
		printk(KERN_ERR "%s: invalid target %d.\n",
			hostdata->name,SCpnt->device->id);
		SCpnt->result = DID_BAD_TARGET << 16;
		SCpnt->scsi_done(SCpnt);
		return 0;
	}

	if (SCpnt->device->lun >= MAX_LUNS &&
			SCpnt->device->channel == AFC_LOGICAL_CHANNEL) {
		SCpnt->result = DID_BAD_TARGET << 16;
		SCpnt->scsi_done(SCpnt);
		return 0;
	}

	if (afc_is_quick_return_command(SCpnt, hostdata))
		return 0;

	spin_lock_irqsave(&afc_scsi_lock, flags);

	/* get free scmd and validate scmd */
	scmd = __get_afc_scmd(&hostdata->free_list);
	if (scmd == NULL) {
		SCpnt->result = (DID_ERROR << 16);
		spin_unlock_irqrestore(&afc_scsi_lock, flags);
		return 1;
	}

	scmd->cmd = SCpnt;
	scmd->message_type = MSG_CDB_EXECUTE;

	/* Set Timeout value to 60 seconds except ioctl */
	if (SCpnt->scsi_done != (void *)ioctl_done &&
			SCpnt->eh_timeout.function)
		mod_timer(&SCpnt->eh_timeout, (jiffies+60*HZ) );

	if (SCpnt->device->channel == AFC_LOGICAL_CHANNEL) {
		if (!(hostdata->pending_flag & (1 << SCpnt->device->id)) ||
				SCpnt->cmnd[0] == REQUEST_SENSE) {
			scmd->state = SCMD_POSTED;
			__do_scsi_queuecommand(SCpnt, hostdata, scmd);
		} else {
			scmd->state = SCMD_PENDING;
			__set_retry_info(SCpnt, hostdata, scmd);
		}
	} else {
		scmd->state = SCMD_POSTED;
		__do_scsi_queuecommand(SCpnt, hostdata, scmd);
	}

	__add_afc_scmd(scmd, &hostdata->exec_list[SCpnt->device->channel]);

	/* statistics */
	hostdata->queue_depth++;
	if (hostdata->queue_depth > hostdata->max_queue_depth) {
		hostdata->max_queue_depth = hostdata->queue_depth;
	}
	hostdata->scsi_cmds++;

	spin_unlock_irqrestore(&afc_scsi_lock, flags);

	return 0;
}


/* Retry queuecommand */
static void __set_retry_info(struct scsi_cmnd * SCpnt,
	struct afc_scsi_host *hostdata, struct afc_scmd *scmd)
{
	if (afc_retry_cnt++ == 0) {
		mod_timer(&afc_retry_timer, jiffies + afc_retry_period);
	}

	hostdata->pending_flag |= 1 << SCpnt->device->id;

	return;
}


static void set_retry_info(struct scsi_cmnd * SCpnt,
	struct afc_scsi_host *hostdata, struct afc_scmd *scmd)
{
	struct afc_target_info	*tinfo;
	unsigned long			flags;

	spin_lock_irqsave(&afc_scsi_lock, flags);
	__set_retry_info(SCpnt, hostdata, scmd);

	tinfo = &hostdata->target_info[SCpnt->device->id];
	tinfo->outgoing_max = afc_outgoing_init;
	tinfo->outgoing_cnt = 0;
	spin_unlock_irqrestore(&afc_scsi_lock, flags);

	return;
}


static struct afc_scmd *__search_pending_scmd(struct afc_scsi_host *hostdata,
	struct list_head *start_lh)
{
	struct list_head	*lh;
	struct afc_scmd		*scmd;

	for (lh = start_lh->prev;
			lh != &hostdata->exec_list[AFC_LOGICAL_CHANNEL]; lh = lh->prev) {
		scmd = list_entry(lh, struct afc_scmd, list);
		if (scmd->state == SCMD_PENDING && scmd->cmd != NULL) {
			return (scmd);
		}
	}

	return (NULL);
}


static void afc_retry_queuecommand(unsigned long f)
{
	struct list_head		*lh;
	struct afc_scsi_host	*hostdata;
	struct afc_scmd			*scmd;
	struct afc_target_info	*tinfo;
	unsigned long			flags;
	u32	retry_flag;
	u8	target;
	int	i,j;

	spin_lock_irqsave(&afc_scsi_lock, flags);

	for (i = 0; i < afc_scsi_hosts; i++) {
		hostdata = hosts[i];
		retry_flag = hostdata->pending_flag;

		/* confirm retry_flag */
		for (j = 0; j < MAX_TARGETS; j++) {
			if (retry_flag & (1 << j)) {
				tinfo = &hostdata->target_info[j];
				if (tinfo->outgoing_max == tinfo->outgoing_cnt)
					retry_flag &= ~(1 << j);
			}
		}

		if (retry_flag == 0)
			continue;

		lh = &hostdata->exec_list[AFC_LOGICAL_CHANNEL];
		while ((scmd = __search_pending_scmd(hostdata, lh)) != NULL) {
			target = scmd->cmd->device->id;

			if (retry_flag & (1 << target)) {
				scmd->state = SCMD_RETRIED;
				__do_scsi_queuecommand(scmd->cmd,hostdata,scmd);
				afc_retry_cnt --;

				tinfo = &hostdata->target_info[target];
				tinfo->outgoing_cnt ++;
				if (tinfo->outgoing_max == tinfo->outgoing_cnt) {
					retry_flag &= ~(1 << target);
					if (retry_flag == 0)
						break;
				}
			}

			lh = &scmd->list;
		}

		hostdata->pending_flag &= ~retry_flag;
	}

	if (afc_retry_cnt) {
		mod_timer(&afc_retry_timer, jiffies + afc_retry_period);
	}

	spin_unlock_irqrestore(&afc_scsi_lock, flags);

	return;
}


/* Statistics */
static void afc_update_lun_stat(unsigned long f)
{
	struct afc_scsi_host	*hostdata;
	struct afc_lun_stat		*c_stat, *p_stat;
	unsigned long			flags;
	int	i,j;

	spin_lock_irqsave(&afc_scsi_lock, flags);

	for (i = 0; i < afc_scsi_hosts; i++) {
		hostdata = hosts[i];
		for (j = 0; j < MAX_TARGETS; j++) {
			c_stat = &hostdata->current_stat[j];
			p_stat = &hostdata->persec_stat[j];
			if (c_stat->r_io_count || c_stat->w_io_count) {
				p_stat->r_io_count =
					c_stat->r_io_count / AFC_STAT_PERIOD;
				p_stat->r_io_size  =
					(c_stat->r_io_size / AFC_STAT_PERIOD) >> 10;
				p_stat->w_io_count =
					c_stat->w_io_count / AFC_STAT_PERIOD;
				p_stat->w_io_size  =
					(c_stat->w_io_size / AFC_STAT_PERIOD) >> 10;
				c_stat->r_io_count = 0;
				c_stat->r_io_size  = 0;
				c_stat->w_io_count = 0;
				c_stat->w_io_size  = 0;
			} else {
				p_stat->r_io_count = 0;
				p_stat->r_io_size  = 0;
				p_stat->w_io_count = 0;
				p_stat->w_io_size  = 0;
			}
		}
	}

	spin_unlock_irqrestore(&afc_scsi_lock, flags);

	mod_timer(&afc_stat_timer, jiffies + AFC_STAT_PERIOD * HZ);
}


/* AFC scsi command */
static void internal_done(struct scsi_cmnd * SCpnt)
{
	SCpnt->SCp.Status++;
	return;
}

int afc_scsi_command(struct scsi_cmnd * SCpnt)
{
	SCpnt->SCp.Status = 0;

	afc_scsi_queuecommand(SCpnt, internal_done);

	while (!SCpnt->SCp.Status)
		barrier();

	return SCpnt->result;
}


/*	AFC abort and reset */
static struct afc_scmd *__search_afc_scmd(struct afc_scsi_host *hostdata,
	struct scsi_cmnd * SCpnt)
{
	struct list_head	*lh;
	struct afc_scmd		*scmd;

	for (lh = hostdata->exec_list[SCpnt->device->channel].prev;
			lh != &hostdata->exec_list[SCpnt->device->channel]; lh = lh->prev) {
		scmd = list_entry(lh, struct afc_scmd, list);
		if (scmd->cmd == SCpnt) {
			return (scmd);
		}
	}

	return (NULL);
}


int afc_scsi_abort(struct scsi_cmnd * SCpnt)
{
	struct Scsi_Host *host = SCpnt->device->host;
	struct afc_scsi_host *hostdata = (struct afc_scsi_host *)host->hostdata;
	struct afc_scmd			*scmd;
	unsigned long			flags;

	spin_lock_irqsave(&afc_scsi_lock, flags);
	scmd = __search_afc_scmd(hostdata, SCpnt);
	spin_unlock_irqrestore(&afc_scsi_lock, flags);
	if (scmd == NULL) {
		printk(KERN_ERR "%s: impossible command to abort.\n",
			hostdata->name);
		return	FAILED;
	}

	printk(KERN_INFO "afxxxc: scsi abort!\n");

	SCpnt->result = (DID_ABORT << 16);
	SCpnt->scsi_done(SCpnt);

	return SUCCESS;
}


int afc_scsi_reset(struct scsi_cmnd * SCpnt)
{
	struct Scsi_Host *host = SCpnt->device->host;
	struct afc_scsi_host *hostdata = (struct afc_scsi_host *)host->hostdata;
	struct afc_scmd				*scmd;
	struct scsi_post_message	*msg;
	unsigned long				flags;
	u32	m;

	/* not reset logical drive */
	if (SCpnt->device->channel == AFC_LOGICAL_CHANNEL) {
		return SUCCESS;
	}

	spin_lock_irqsave(&afc_scsi_lock, flags);

	/* test and set reset flag */
	if (hostdata->reset_flag & (0x1<<SCpnt->device->channel)) {
		spin_unlock_irqrestore(&afc_scsi_lock, flags);
		return FAILED;
	}

	hostdata->reset_flag |= 0x1<<SCpnt->device->channel;

	/* get free scmd */
	scmd = __get_afc_scmd(&hostdata->free_list);
	if (scmd == NULL) {
		spin_unlock_irqrestore(&afc_scsi_lock, flags);
		printk(KERN_ERR "%s: unable to get scmd in scsi reset!!\n",
			hostdata->name);
		return FAILED;
	}

	/* get inbound message */
	m = AFC_POST_READ32(hostdata);
	if (m == 0xFFFFFFFF) {
		__add_afc_scmd(scmd, &hostdata->free_list);
		spin_unlock_irqrestore(&afc_scsi_lock, flags);
		printk(KERN_ERR "%s: unable to get msg in scsi reset!!\n",
			hostdata->name);
		return FAILED;
	}
	msg = (struct scsi_post_message *)(hostdata->mem_offset + m);

	/* set scmd */
	scmd->state = SCMD_POSTED;
	scmd->cmd = SCpnt;
	scmd->message_type = MSG_BUS_RESET;

	/* set header */
	if (SCpnt->device->channel == AFC_LOGICAL_CHANNEL)
		make_header(SCpnt, hostdata, msg, scmd, MSG_BUS_DEV_RESET);
	else
		make_header(SCpnt, hostdata, msg, scmd, MSG_BUS_RESET);

	__afc_scsi_post_message(hostdata, scmd, m, SCpnt->device->channel);

	spin_unlock_irqrestore(&afc_scsi_lock, flags);

	printk(KERN_INFO "afxxxc: scsi reset!\n");

	return SUCCESS;
}


/* Return the disk geometry for a particular disk */
int afc_scsi_bios_param(struct scsi_device *sdev, struct block_device *bdev ,
	sector_t cap, int *ip)
{
	int size;

	size = cap;

	ip[0] = 64;				/* heads */
	ip[1] = 32;				/* sectors */
	ip[2] = size >> 11;			/* cylinders */

	if (ip[2] > 1024) {			/* for big disk */
		ip[0] = 255;			/* heads */
		ip[1] = 63;			/* sectors */
		ip[2] = size / (255 * 63);	/* cylinders */
	}

	return 0;
}


#ifdef AFC_64BIT_PTR
u64 linux_get_memsize(void)
{
	struct sysinfo	si;

	si_meminfo(&si);
	return ((u64)si.totalram << PAGE_SHIFT);
}
#endif

static void return_to_bios_mode(struct afc_scsi_host *hostdata, u8 *mem)
{
	struct scsi_cmnd			*SCpnt;
	struct scsi_device			*SDpnt;
	struct request				*req;
	struct afc_scmd				*scmd;
	struct scsi_post_message	*msg;
	u32							m;

	/* Get Scsi Request Buffer */
	SDpnt = (struct scsi_device *)
			kmalloc(sizeof(struct scsi_device), GFP_KERNEL);
	if (!SDpnt) {
		return;
	}
	memset(SDpnt, 0, sizeof(struct scsi_device));

	SCpnt = (struct scsi_cmnd *)kmalloc(sizeof(struct scsi_cmnd), GFP_KERNEL);
	if (!SCpnt) {
		kfree(SDpnt);
		return;
	}
	memset(SCpnt, 0, sizeof(struct scsi_cmnd));

	req = (struct request *)kmalloc(sizeof(struct request), GFP_KERNEL);
	if (!req) {
		kfree(SCpnt);
		kfree(SDpnt);
		return;
	}
	memset(req, 0, sizeof(struct request));

	SCpnt->device		= SDpnt;
	SCpnt->device->host	= hostdata->host;
	SCpnt->device->channel	= 0;
	SCpnt->device->id	= 0;
	SCpnt->device->lun	= 0;
	SCpnt->request		= req;
	SCpnt->request->waiting	= 0;
	SCpnt->use_sg		= 0;
	SCpnt->scsi_done	= NULL;

	/* get inbound message */
	m = AFC_POST_READ32(hostdata);
	if (m == 0xFFFFFFFF) {
		printk(KERN_ERR "%s: unable to return to bios mode\n",
		 hostdata->name);
		kfree(req);
		kfree(SCpnt);
		kfree(SDpnt);
		return;
	}
	msg = (struct scsi_post_message *)(hostdata->mem_offset + m);

	/* get free scmd */
	scmd = get_afc_scmd(&hostdata->free_list);
	if (scmd == NULL) {
		printk(KERN_ERR "%s: unable to return to bios mode\n",
		 hostdata->name);
		kfree(req);
		kfree(SCpnt);
		kfree(SDpnt);
		return;
	}

	/* set scmd */
	scmd->state = SCMD_POSTED;
	scmd->cmd = SCpnt;
	scmd->message_type = MSG_RST_O_FREE_QUE;

	/* set header */
	make_header(SCpnt, hostdata, msg, scmd, MSG_RST_O_FREE_QUE);

	afc_scsi_post_message(hostdata, scmd, m, 0);

	/* waiting that FW returns to BIOS mode */
	while (1) {
		if (readl(mem + IOP348_ATU_OMWTVR_L) == 0xFFFFFFFF)
			break;
	}

	__free_afc_scmd(hostdata, scmd);

	kfree(req);
	kfree(SCpnt);
	kfree(SDpnt);

	return;
}

static int __devinit afc_probe_one(struct pci_dev *pdev,
	const struct pci_device_id *id)
{
	struct afc_scsi_host	*hostdata;
	int		devnum = id->driver_data;
	struct Scsi_Host	*shpnt;
	struct scsi_reply_message	*mbuf;
	unsigned long	memptr = 0, size;
	dma_addr_t	mbuf_phys, mbuf_bus;
	u8		*mem;
	int		i;
	int		error = -ENODEV;
#ifdef AFC_64BIT_PTR
	u64		memsize;
	const u64 mask_40bit = 0xFFFFFFFFFFULL;
#endif
	u32		bios_command_reg_value;
	u32		mbus;
int	inCQ = 0;

	/* detected a new controller */
	printk(KERN_INFO
		"afxxxc: probe new device %#4.04x:%#4.04x:%#4.04x:%#4.04x\n",
			pdev->vendor, pdev->device,
			pdev->subsystem_vendor, pdev->subsystem_device);

	printk(KERN_INFO "afxxxc: bus %d:slot %d:func %d\n", pdev->bus->number,
			PCI_SLOT(pdev->devfn), PCI_FUNC(pdev->devfn));

	/* Enable bus-mastering on this controller */
	if (pci_enable_device(pdev)) {
		printk(KERN_ERR
			"afxxxc: Failed to enabled pci device, aborting.\n");
		return error;
	}
	pci_set_master(pdev);

	/* Skip I/O spaces */
	for (i = 0; i < 6; i++) {
		if (!(pci_resource_flags(pdev, i) & IORESOURCE_IO)) {
			memptr = pci_resource_start(pdev, i);
			break;
		}
	}
	if (i == 6) {
		printk(KERN_ERR
			"afxxxc: hostdata has no memory regions defined.\n");
		error = -EINVAL;
		goto error_disable_device;
	}

	if (pci_request_regions(pdev, "TOSHIBA AFC SCSI DRIVER") != 0) {
		printk(KERN_ERR
			"afxxxc: hostdata has no memory regions defined.\n");
		error = -EINVAL;
		goto error_disable_device;
	}

	/* map the AFC host adapter */
	size = AFC_MEM_SIZE;
	mem = ioremap_nocache(memptr, size);
	if (mem == NULL) {
		printk(KERN_ERR "afxxxc: Unable to map controller.\n");
		error = -EINVAL;
		goto error_release_region;
	}

	/* setup the DMA mask */
#ifdef AFC_64BIT_PTR
	memsize = linux_get_memsize();
	if (memsize >= 0x10000000000ULL) {
		printk(KERN_WARNING
			"afxxxc: Unable to set a suitable DMA mask - aborting\n");
		error = -ENODEV;
		goto error_iounmap;
	}
	if (memsize < 0x10000000000ULL) {
		if (pci_set_dma_mask(pdev, mask_40bit)) {
			if (pci_set_dma_mask(pdev, DMA_32BIT_MASK)) {
				printk(KERN_WARNING
		"afxxxc: Unable to set a suitable DMA mask - aborting\n");
				error = -ENODEV;
				goto error_iounmap;
			}
		}
	}
#else
	if (pci_set_dma_mask(pdev, DMA_32BIT_MASK)) {
		printk(KERN_WARNING
			"afxxxc: Unable to set a suitable DMA mask - aborting\n");
		error = -ENODEV;
		goto error_iounmap;
	}
#endif

	/* register scsi host adapter */
	shpnt = scsi_host_alloc(&AFCSCSI, sizeof(struct afc_scsi_host));
	if (!shpnt) {
		printk(KERN_ERR "afxxxc: Failed to register host, aborting.\n");
		error = -ENOMEM;
		goto error_iounmap;
	}
#ifdef PROPRIETARY_ERROR_HANDLER
	shpnt->transportt = &AFC_TRANSPORT;
#endif
	hostdata = (struct afc_scsi_host *)shpnt->hostdata;
	memset(hostdata, 0, sizeof(struct afc_scsi_host));

	/* physical resoureces */
	hostdata->pci.dev	= pdev;
	hostdata->devnum	= devnum;
	hostdata->pci.did	= pdev->device;
	hostdata->pci.sdid	= pdev->subsystem_device;
	hostdata->pci.irq	= pdev->irq;
	hostdata->irq_mask	= (u32 *)(mem + 0x34);
	hostdata->post_port	= (u32 *)(mem + 0x40);
	hostdata->reply_port	= (u32 *)(mem + 0x44);
	hostdata->mem_phys	= memptr;
	hostdata->mem_offset	= (unsigned long)mem;

	/* disable interrupt */
	AFC_IRQ_WRITE32(hostdata, 0xFFFFFFFF);
	barrier();

	/* Check if we are in CQ mode or not */
	switch (afc_board_tbl[hostdata->devnum].generation) {
	case 3:
		pci_read_config_dword(
			hostdata->pci.dev, IOP348_BIOS_COMMAND_REG,
			&bios_command_reg_value);
		break;
	case 2:
		pci_read_config_dword(
			hostdata->pci.dev, IOP33X_BIOS_COMMAND_REG,
			&bios_command_reg_value);
		break;
	default:
		bios_command_reg_value = 0;
	}

	if ((bios_command_reg_value & DISABLE_BIOS_ACCESS) ||
		(bios_command_reg_value & LINUX_MODE_ENABLE)) {

		inCQ = 1;
		/* wait for terminating i/o */
		msleep(30*HZ);

		/* clear interrupt */
		while ((mbus = AFC_REPLY_READ32(hostdata)) != 0xFFFFFFFF)
		{
			afc_flush_reply(hostdata,mbus);
			mb();
		}
	}

	/* install interrupt handler */
	if (request_irq(hostdata->pci.irq, afc_pci_interrupt, SA_SHIRQ,
			hostdata->name, hostdata)) {
		printk(KERN_ERR
			"afxxxc: Unable to allocate interrupt (%d).\n", pdev->irq);
		error = -EBUSY;
		goto error_scsi_put_host;
	}

	/* initialize retry timer */
	if (afc_scsi_hosts == 0) {
		init_timer(&afc_retry_timer);
		afc_retry_timer.data = 0UL;
		afc_retry_timer.function = afc_retry_queuecommand;

		/* initialize stat timer */
		init_timer(&afc_stat_timer);
		afc_stat_timer.expires = jiffies + AFC_STAT_PERIOD * HZ;
		afc_stat_timer.function = afc_update_lun_stat;
		afc_stat_timer.data = 0UL;
		add_timer(&afc_stat_timer);
	}

	/* initialize scmd list */
	for (i = 0; i <= afc_board_tbl[devnum].numPorts; i++) {
		INIT_LIST_HEAD(&hostdata->exec_list[i]);
	}
	INIT_LIST_HEAD(&hostdata->free_list);

	for (i = 0; i < MAX_COMMANDS; i++) {
		hostdata->scmd[i].tag_no = i;
		__add_afc_scmd(&hostdata->scmd[i], &hostdata->free_list);
	}

	if ( inCQ ) {
		if (afc_board_tbl[hostdata->devnum].generation >= 2) {
			return_to_bios_mode(hostdata, mem);
		}
	}
	/* initialize outbound message buffer */
	mbuf = allocate_reply_buffer(pdev, &mbuf_bus);

	if (mbuf == NULL) {
		printk(KERN_ERR "afxxxc: unable to allocate message buffer.\n");
		error = -EINVAL;
		goto error_free_irq_host;
	}
	memset(mbuf, (int)0xFFFFFFFF, AFC_REPLY_SIZE * AFC_REPLY_FRAMES);

	hostdata->page_frame = mbuf;
	hostdata->page_frame_bus = mbuf_bus;

	switch (afc_board_tbl[hostdata->devnum].generation) {
	case 3:
		writel((u32)((u64)mbuf_bus >> 32), mem + IOP348_ATU_OMWTVR_H);
		writel((u32)((u64)mbuf_bus & 0x00000000FC000000ull),
			mem + IOP348_ATU_OMWTVR_L);

		hostdata->csr_phys = 0;
		hostdata->csr_offset = 0;
		barrier();

		/* enter CQ mode with 'DISABLE_BIOS_ACCESS' */
		pci_write_config_dword(
			hostdata->pci.dev, IOP348_BIOS_COMMAND_REG,
			DISABLE_BIOS_ACCESS);

		/* to set linux mode again, once back to bios mode */
		return_to_bios_mode(hostdata, mem);

		/* set reply frames */
		for (i = 0; i < AFC_REPLY_FRAMES; i ++) {
			mbuf_phys = (((mbuf_bus + i * sizeof(struct scsi_reply_message))
				& 0x03FFFFFF) | 0x80000000);
			AFC_REPLY_WRITE32(hostdata, mbuf_phys);
		}

		writel((u32)((u64)mbuf_bus >> 32), mem + IOP348_ATU_OMWTVR_H);
		writel((u32)((u64)mbuf_bus & 0x00000000FC000000ull),
			mem + IOP348_ATU_OMWTVR_L);

		/* set linux auto sense mode */
		pci_write_config_dword(
			hostdata->pci.dev, IOP348_BIOS_COMMAND_REG,
			LINUX_MODE_ENABLE);
		break;
	case 2:
		pci_write_config_dword(hostdata->pci.dev,
			IOP33X_ATU_OUMWTVR0,(u32)((u64)mbuf_bus >> 32));
		pci_write_config_dword(hostdata->pci.dev,
			IOP33X_ATU_OMWTVR0,(u32)((u64)mbuf_bus & 0xFC000000ull));

		hostdata->csr_phys = 0;
		hostdata->csr_offset = 0;
		barrier();

		/* set reply frames */
		for (i = 0; i < AFC_REPLY_FRAMES; i ++) {
			mbuf_phys = (((mbuf_bus + i * sizeof(struct scsi_reply_message))
				& 0x03FFFFFF) | 0x80000000);
			AFC_REPLY_WRITE32(hostdata, mbuf_phys);
		}

		pci_write_config_dword(hostdata->pci.dev,
			IOP33X_BIOS_COMMAND_REG,
			LINUX_MODE_ENABLE);
		break;
	default:
		hostdata->csr_phys = 0;
		hostdata->csr_offset = 0;
		barrier();
		break;
	}

	/* enable interrupt */
	AFC_IRQ_WRITE32(hostdata, 7);

	/* setup adapter handle in PCI soft state */
	pci_set_drvdata(hostdata->pci.dev, shpnt);

	/* attach with scsi mid-layer */
	hostdata->host = shpnt;
	hostdata->host_no = shpnt->host_no;
	sprintf(hostdata->name, "afxxxc%d", afc_scsi_hosts);

	shpnt->max_channel	= afc_board_tbl[devnum].numPorts;
	shpnt->max_id		= MAX_TARGETS;
	shpnt->can_queue	= afc_board_tbl[devnum].can_queue;
	shpnt->cmd_per_lun	= afc_board_tbl[devnum].cmd_par_lun;

	error = scsi_add_host(shpnt, &hostdata->pci.dev->dev);

	if (error)
		goto error_disable_adapter;

	/* set writeback cache on */
	if ( !inCQ ) {
		afc_cache_mode(hostdata, PAGE3C_PNP_START_ADAPTER);
		afc_cache_mode(hostdata, PAGE3C_CACHE_ON);

		barrier();
	}

	/* register miscdevice for iotcl */
	if ( afc_scsi_hosts == 0 )
		afc_install_ioctl();

	scsi_scan_host(shpnt);

	hosts[afc_scsi_hosts] = hostdata;
	afc_scsi_hosts ++;

	return 0;

error_disable_adapter:
	pci_set_drvdata(hostdata->pci.dev, NULL);
        switch (afc_board_tbl[hostdata->devnum].generation) {
        case 3:
                pci_read_config_dword(
                        hostdata->pci.dev, IOP348_BIOS_COMMAND_REG,
                        &bios_command_reg_value);
                pci_write_config_dword(
                        hostdata->pci.dev, IOP348_BIOS_COMMAND_REG,
                        (bios_command_reg_value &
                        ~(DISABLE_BIOS_ACCESS | LINUX_MODE_ENABLE)));
                break;
        case 2:
                pci_read_config_dword(
                        hostdata->pci.dev, IOP33X_BIOS_COMMAND_REG,
                        &bios_command_reg_value);
                pci_write_config_dword(
                        hostdata->pci.dev, IOP33X_BIOS_COMMAND_REG,
                        (bios_command_reg_value & ~LINUX_MODE_ENABLE));
                break;
        default:
                bios_command_reg_value = 0;
                break;
	}
	dma_free_coherent(&pdev->dev, (AFC_REPLY_SIZE * AFC_REPLY_FRAMES),
		(void *)hostdata->page_frame, hostdata->page_frame_bus);

error_free_irq_host:
	del_timer_sync(&afc_stat_timer);
	del_timer_sync(&afc_retry_timer);
	free_irq(hostdata->pci.irq, hostdata);

error_scsi_put_host:
	scsi_host_put(shpnt);

error_iounmap:
	iounmap(mem);

error_release_region:
	pci_release_regions(pdev);

error_disable_device:
	pci_disable_device(pdev);

	return error;
}


#if LINUX_VERSION_CODE < LinuxVersionCode(2,6,15)
static void afc_shutdown(struct device *dev)
{
	struct Scsi_Host	*host = pci_get_drvdata(to_pci_dev(dev));
#else
static void afc_shutdown(struct pci_dev *pdev)
{
	struct Scsi_Host	*host = pci_get_drvdata(pdev);
#endif
	struct afc_scsi_host	*hostdata =
		(struct afc_scsi_host *)host->hostdata;

	printk(KERN_INFO "afxxxc: Shutting down AFC SCSI system.\n");

	afc_cache_mode(hostdata, PAGE3C_CACHE_OFF);
	afc_cache_mode(hostdata, PAGE3C_PNP_STOP_ADAPTER);
	afc_cache_mode(hostdata, PAGE3C_SHUTDOWN);
	flush_outbound_message(hostdata);

	printk(KERN_INFO "afxxxc: AFC SCSI system down.\n");
	return;
}


static void __devexit afc_remove_one(struct pci_dev *pdev)
{
	struct Scsi_Host	*shpnt = pci_get_drvdata(pdev);
	struct afc_scsi_host	*hostdata =
		(struct afc_scsi_host *)shpnt->hostdata;
	int	i;
	u32     bios_command_reg_value;

	afc_cache_mode(hostdata, PAGE3C_CACHE_OFF);
	afc_cache_mode(hostdata, PAGE3C_PNP_STOP_ADAPTER);

	scsi_remove_host(shpnt);

	flush_outbound_message(hostdata);

	pci_set_drvdata(pdev, NULL);

	switch (afc_board_tbl[hostdata->devnum].generation) {
	case 3:
		pci_read_config_dword(
			hostdata->pci.dev, IOP348_BIOS_COMMAND_REG,
			&bios_command_reg_value);
		pci_write_config_dword(
			hostdata->pci.dev, IOP348_BIOS_COMMAND_REG,
			(bios_command_reg_value &
			~(DISABLE_BIOS_ACCESS | LINUX_MODE_ENABLE)));
		break;
	case 2:
		pci_read_config_dword(
			hostdata->pci.dev, IOP33X_BIOS_COMMAND_REG,
			&bios_command_reg_value);
		pci_write_config_dword(
			hostdata->pci.dev, IOP33X_BIOS_COMMAND_REG,
			(bios_command_reg_value & ~LINUX_MODE_ENABLE));
		break;
	default:
		bios_command_reg_value = 0;
		break;
	}

	dma_free_coherent(&pdev->dev, (AFC_REPLY_SIZE * AFC_REPLY_FRAMES),
		(void *)hostdata->page_frame, hostdata->page_frame_bus);

	if (afc_scsi_hosts == 1)
		del_timer_sync(&afc_stat_timer);

	/* disable interrupt */
	AFC_IRQ_WRITE32(hostdata, 0xFFFFFFFF);
	barrier();

	return_to_bios_mode (hostdata, (u8*)(hostdata->post_port)-0x40);

	free_irq(hostdata->pci.irq, hostdata);

	if (--afc_scsi_hosts == 0) {
		del_timer_sync(&afc_retry_timer);
		afc_remove_ioctl();
	}

	for (i = 0; i < afc_hosts_max; i++) {
		if (hosts[i] == hostdata) {
			hosts[i] = NULL;
			break;
		}
	}

	scsi_host_put(shpnt);

	iounmap(((u8 *)hostdata->post_port) - 0x40);

	pci_release_regions(pdev);

	pci_disable_device(pdev);

	printk(KERN_INFO "afxxxc: removed AFC SCSI device driver.\n");

	return;
}


static int __init afc_init(void)
{
	printk(KERN_NOTICE
		"afxxxc: TOSHIBA RAID controller device driver Ver.%s\n", AFC_VERSION);

	return (pci_register_driver(&afc_pci_driver));
}


static void __exit afc_exit(void)
{
	printk(KERN_NOTICE "afxxxc: unloading driver.\n");

	pci_unregister_driver(&afc_pci_driver);

	return;
}

MODULE_AUTHOR("Toshiba Solutions Corporation");
MODULE_DESCRIPTION("AFC SCSI driver");
MODULE_LICENSE("GPL");
MODULE_VERSION(AFC_VERSION);
module_param(afc_hosts_max,		int, 0644);
module_param(afc_outgoing_init,		int, 0644);
module_param(afc_retry_period,		int, 0644);
module_param(afc_ioctl_timeout,		int, 0644);
module_param(afc_ioctl_retry_max,	int, 0644);

module_init(afc_init);
module_exit(afc_exit);
