/*
 *  mazedev.c, Maze monitoring function driver.
 *
 *  Copyright (C) 2007-2008 MIRACLE LINUX Corp.
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/hash.h>
#include <linux/time.h>
#include <linux/timer.h>
#include <linux/poll.h>
#include <linux/version.h>
#include <linux/cdev.h>
#include <linux/types.h>
#include <linux/utrace.h>
#include <linux/uaccess.h>
#include <linux/delay.h>
#include <linux/unistd.h>
#include <linux/ptrace.h>

#include "maze.h"
#include "mazequeue.h"
#include "mazetimer.h"

/*
 *  Maze trace events for utrace
 */
#define MAZE_EVENTS (UTRACE_EVENT(EXIT) | \
					UTRACE_EVENT(CLONE) | \
					UTRACE_EVENT(SYSCALL_ENTRY) | \
					UTRACE_EVENT(SYSCALL_EXIT))


/* Maze hash table data */
struct hlist_head mz_hash[MZHASH_NR];

/* Protects mz_hash */
spinlock_t mzhlist_lock[MZHASH_NR];

/* Notice data */
LIST_HEAD(notice_buf_list);

/* Protects notice_buf_list */
spinlock_t notice_buf_lock;
int notice_num;

/* Wait queue head for polling maze device */
DECLARE_WAIT_QUEUE_HEAD(maze_wait);

/*
 *  Default values
 */
int soft_lim_default = MAZE_LIM_SOFTDEF;
int hard_lim_default = MAZE_LIM_HARDDEF;
int soft_sig_default = MAZE_SIG_SOFTDEF;
int hard_sig_default = MAZE_SIG_HARDDEF;
int maze_attr_flag;

/*
 *  Module parameters
 */
int eip_disp_count;
int use_notice;
int notice_buf_count = 16;
int mazedev_major;
int maze_timer_interval = 10;
module_param(eip_disp_count, int, S_IRUGO);
module_param(use_notice, bool, S_IRUGO|S_IWUSR);
module_param(notice_buf_count, int, S_IRUGO);
module_param(mazedev_major, int, S_IRUGO);
module_param(maze_timer_interval, int , S_IRUGO);

/*
 *  Prototypes for functions
 */
extern int  mzdev_install_proc(void);
extern void mzdev_remove_proc(void);

/*
 * data for device driver registraion
 */
static dev_t maze_devno;
static struct cdev *maze_cdev;

/* statistical counts */
static atomic_t mzdev_open_count;
static atomic_t mzdev_release_count;


static u32 syscall_entry_mazecnt(struct utrace_attached_engine *engine,
								 struct task_struct *task,
								 struct pt_regs *regs);

static u32 syscall_exit_mazecnt(struct utrace_attached_engine *engine,
								struct task_struct *task,
								struct pt_regs *regs);

static u32 clone_mazeinfo(struct utrace_attached_engine *engine,
						  struct task_struct *parent,
						  unsigned long clone_flags,
						  struct task_struct *child);

static u32 exit_mazeinfo(struct utrace_attached_engine *engine,
						 struct task_struct *task,
						 long orig_code, long *code);

static struct utrace_engine_ops task_ops = {
	.report_clone = clone_mazeinfo,
	.report_exit = exit_mazeinfo,
	.report_syscall_entry = syscall_entry_mazecnt,
	.report_syscall_exit = syscall_exit_mazecnt,
};

/*
 *  allocate to maze_notice
 */
static inline struct notice_entry *alloc_notice_entry(void)
{
	return kmalloc(sizeof(struct notice_entry), GFP_ATOMIC);
}

/*
 *  find specified entry from the mz_hash table
 * Note. Must be called rcu_read_lock().
 */
static struct maze_entry *fastcall find_maze_entry(int nr)
{
	struct hlist_node *elem;
	struct maze_entry *entry;

	hlist_for_each_entry_rcu(entry, elem, &mz_hash[mz_hashfn(nr)], chain) {
		if (entry->pid == nr)
			return entry;
	}

	return NULL;
}

static void timer_handler(unsigned long data);

/*
 *  Add to the mz_hash table
 */
static int add_maze_entry(struct maze_data *data, struct task_struct *task)
{
	struct maze_queue_entry *qentry;
	struct maze_entry *entry;
	struct utrace_attached_engine *engine;
	unsigned long flags;
	int idx;

	set_cpu_timer(timer_handler, maze_timer_interval);

	engine = utrace_attach(task, UTRACE_ATTACH_CREATE , &task_ops, 0);
	if (IS_ERR_VALUE((long)engine)) {
		ERRPRINT("fail to attach utrace\n");
		return -PTR_ERR(engine);
	}

	qentry = alloc_queue_entry();
	if (unlikely(!qentry)) {
		utrace_detach(task, engine);
		ERRPRINT("fail to alloc maze_queue_entry\n");
		return -ENOMEM;
	}

	entry = maze_entry(qentry);
	entry->pid = data->pid;
	entry->soft_lim = data->soft_lim;
	entry->hard_lim = data->hard_lim;
	entry->soft_sig = data->soft_sig;
	entry->hard_sig = data->hard_sig;
	entry->sig_flag = NOT_SEND_SIG;
	entry->flag = data->flag;
	entry->nivcsw = task->nivcsw;
	entry->nvcsw = task->nvcsw;
	entry->reset_utime = task->utime;
	entry->reset_stime = task->stime;

	idx = mz_hashfn(entry->pid);
	spin_lock_irqsave(&mzhlist_lock[idx], flags);
	hlist_add_head_rcu(&entry->chain, &mz_hash[idx]);
	enqueue(qentry);
	spin_unlock_irqrestore(&mzhlist_lock[idx], flags);

	utrace_set_flags(task, engine, MAZE_EVENTS);

	DBGPRINT("Add maze entry for pid %d in hash %ld\n",
			 entry->pid, mz_hashfn(entry->pid));
	return 0;
}

/*
 *  Search specified maze entry and remove it
 */
static int del_maze_entry(struct task_struct *task,
						  struct utrace_attached_engine *engine)
{
	struct maze_queue_entry *qentry;
	struct maze_entry *ment;
	unsigned long flags;
	int idx;

	set_cpu_timer(timer_handler, maze_timer_interval);

	idx = mz_hashfn(task->pid);
	spin_lock_irqsave(&mzhlist_lock[idx], flags);
	utrace_detach(task, engine);
	ment = find_maze_entry(task->pid);

	if (likely(ment)) {
		hlist_del_rcu(&ment->chain);

		qentry = maze_queue_entry(ment);
		dequeue_and_free(qentry);

		spin_unlock_irqrestore(&mzhlist_lock[idx], flags);
		DBGPRINT("Remove maze entry for pid %d\n", task->pid);
		return 0;
	}
	spin_unlock_irqrestore(&mzhlist_lock[idx], flags);

	return -1;
}

/*
 * Find maze_entry by pid and copy maze_data
 * Note. Must be called rcu_read_lock()
 */
static int set_maze_entry(struct maze_data *data, struct task_struct *task)
{
	int ret = 0;
	struct maze_entry *entry;

	entry = find_maze_entry(data->pid);
	if (entry == NULL) {
		ret = add_maze_entry(data, task);
	} else {
		/* We don't reset the member "count" */
		entry->soft_lim = data->soft_lim;
		entry->hard_lim = data->hard_lim;
		entry->soft_sig = data->soft_sig;
		entry->hard_sig = data->hard_sig;
		entry->flag = data->flag;
	}

	return ret;
}

/*
 *  Get first notice entry from the buffer
 */
static struct notice_entry *get_notice_entry(void)
{
	if (list_empty(&notice_buf_list))
		return NULL;

	/* point first notice entry and return */
	return list_entry(notice_buf_list.next, struct notice_entry, chain);
}

/*
 *  Delete specified notice entry from the buffer
 */
static void del_notice_entry(struct notice_entry *p)
{
	/* delete the entry */
	list_del(&(p->chain));
	kfree(p);
	notice_num--;
}

/*
 *  Allocate new notice entry and add to the notice buffer list
 */
static int add_notice_entry(struct maze_notice *mnp)
{
	struct notice_entry *p;
	struct notice_entry *del;
	unsigned long flags;

	p = alloc_notice_entry();
	if (!p) {
		ERRPRINT("fail to alloc mn_entry\n");
		return -ENOMEM;
	}

	p->mn = *mnp;

	/*
	 * Delete oldest notice data if notice buffer is full
	 */
	spin_lock_irqsave(&notice_buf_lock, flags);
	if (notice_buf_count && notice_num >= notice_buf_count) {
		del = get_notice_entry();
		if (!del)
			panic("notice_num indicates buffer is FULL,"
			      " but no entry can be found!\n");

		del_notice_entry(del);
		DBGPRINT("Notice Buffer is FULL and oldest data is over-written\n");
	}

	/*
	 * Add new entry and update counter
	 */
	list_add_tail(&(p->chain), &notice_buf_list);
	notice_num++;
	spin_unlock_irqrestore(&notice_buf_lock, flags);

	return 0;
}

/*
 *  Setup notice data and add to the buffer
 */
static void do_notice(int pid, int event, int data1, int data2)
{
	struct maze_notice notice;
	int ret;

	notice.pid = pid;
	notice.event = event;
	do_gettimeofday(&notice.tv);
	notice.data1 = data1;
	notice.data2 = data2;

	ret = add_notice_entry(&notice);
	if (ret) {
		ERRPRINT("notice data1 add error : "
				 "pid %d, event %d, data1 %d, data2 %d\n",
				 pid, event, data1, data2);
		return;
	}

	wake_up_interruptible(&maze_wait);
}

/*
 *  Check if process inheritance flag is ON.
 *  If YES then return 1, otherwise return 0.
 */
static inline int inhr_process(int flag)
{
	if ((maze_attr_flag & MAZE_ATTR_INHRCHILD) ||
	   (flag & MAZE_ATTR_INHRCHILD))
		return 1;

	return 0;
}

/*
 *  Check if thread inheritance flag is ON.
 *  If YES then return 1, otherwise return 0.
 */
static inline int inhr_thread(int flag)
{
	if ((maze_attr_flag & MAZE_ATTR_INHRTHREAD) ||
	   (flag & MAZE_ATTR_INHRTHREAD))
		return 1;

	return 0;
}

/*
 *  Check if monitor all flag is ON.
 *  If YES then return 1, otherwise return 0.
 */
static inline int mon_all(void)
{
	if (maze_attr_flag & MAZE_ATTR_MONITORALL)
		return 1;

	return 0;
}

static inline void set_clone_mazeinfo(struct maze_data *new, struct maze_entry *old)
{
	new->soft_lim = old->soft_lim;
	new->hard_lim = old->hard_lim;
	new->soft_sig = old->soft_sig;
	new->hard_sig = old->hard_sig;
	new->flag = old->flag;
}

/*
 *  Clone maze monitoring data
 */
static u32 clone_mazeinfo(struct utrace_attached_engine *engine,
						  struct task_struct *parent,
						  unsigned long clone_flags,
						  struct task_struct *child)
{
	struct maze_entry *ment;
	struct maze_data new;
	int ret;

	rcu_read_lock();
	ment = find_maze_entry(child->pid);
	if (unlikely(ment))
		panic("Maze data for PID %d is already registered\n", child->pid);

	/* Search parrent maze info data */
	if (!(clone_flags & CLONE_THREAD)) {
		/* Find Parent's maze entry */
		ment = find_maze_entry(parent->pid);
		if (ment) {
			if (inhr_process(ment->flag)) {
				new.pid = child->pid;
				set_clone_mazeinfo(&new, ment);
				rcu_read_unlock();
				ret = add_maze_entry(&new, child);
				if (unlikely(ret))
					ERRPRINT("Can't create a maze entry for new process : "
							 "PID %d\n", child->pid);
				return 0;
			}
		}
	} else {
		/* Find Thead Group Leader's maze entry */
		ment = find_maze_entry(parent->tgid);
		if (ment) {
			if (inhr_thread(ment->flag)) {
				new.pid = child->pid;
				set_clone_mazeinfo(&new, ment);
				rcu_read_unlock();
				ret = add_maze_entry(&new, child);
				if (unlikely(ret))
					ERRPRINT("Can't create a maze entry for new thread : "
							 "PID %d, TGID %d\n", child->pid, current->tgid);
				return 0;
			}
		}
	}
	rcu_read_unlock();

	/*
	 *  Create a maze entry with default values if ALL Monitor flag is ON
	 */
	if (mon_all()) {
		new.pid = child->pid;
		new.soft_lim = soft_lim_default;
		new.hard_lim = hard_lim_default;
		new.soft_sig = soft_sig_default;
		new.hard_sig = hard_sig_default;
		new.flag = 0;

		ret = add_maze_entry(&new, child);
		if (ret)
			ERRPRINT("Can't create a maze entry for PID %d \n", child->pid);
		return 0;
	}

	return 0;
}

/*
 *  Remove maze monitoring data
 */
static u32 exit_mazeinfo(struct utrace_attached_engine *engine,
						 struct task_struct *task,
						 long orig_code, long *code)
{
	int ret = del_maze_entry(task, engine);

	if (ret) {
		DBGPRINT("already removed a maze info data of PID %d \n", task->pid);
		return 0;
	}

	DBGPRINT("maze info data of PID %d removed\n", task->pid);
	if (use_notice)
		do_notice(task->pid, MAZE_EV_EXITTARGET, 0, 0);

	return 0;
}

/*
 * get monitoring process states
 */
static u32 syscall_entry_mazecnt(struct utrace_attached_engine *engine,
								 struct task_struct *task,
								 struct pt_regs *regs)
{
	struct maze_entry *mzent;

	set_cpu_timer(timer_handler, maze_timer_interval);
	rcu_read_lock();
	mzent = find_maze_entry(task->pid);
	if (likely(mzent)) {
		mzent->nvcsw = task->nvcsw;
		mzent->nivcsw = task->nivcsw;
	}
	rcu_read_unlock();

	return 0;
}

/*
 * check context switch counter
 */
static u32 syscall_exit_mazecnt(struct utrace_attached_engine *engine,
								struct task_struct *task,
								struct pt_regs *regs)
{
	struct maze_entry *mzent;

	set_cpu_timer(timer_handler, maze_timer_interval);
	rcu_read_lock();
	mzent = find_maze_entry(task->pid);
	if (likely(mzent)) {
		if (mzent->nvcsw != task->nvcsw ||
			__NR_sched_yield == *tracehook_syscall_callno(regs)){
			mzent->reset_utime = task->utime;
			mzent->reset_stime = task->stime;
			if (mzent->sig_flag & SENT_SOFT_SIG)
				mzent->flag = NOT_SEND_SIG;

		}
		might_sleep_list(maze_queue_entry(mzent), task);
	}
	rcu_read_unlock();

	return 0;
}

/*
 * Get soft limit value
 */
static inline int get_soft_limit(struct maze_entry *entry)
{
	return entry->soft_lim == MAZE_LIM_USEDEFAULT ?
		soft_lim_default : entry->soft_lim;
}

/*
 * Get hard limit value
 */
static inline int get_hard_limit(struct maze_entry *entry)
{
	return entry->hard_lim == MAZE_LIM_USEDEFAULT ?
		hard_lim_default : entry->hard_lim;
}

static inline cputime_t get_compare_value(struct maze_queue_entry *qentry)
{
	struct maze_entry *entry = maze_entry(qentry);
	struct task_struct *task = find_task_by_pid(entry->pid);
	cputime_t ret;

	if (unlikely(!task))
		return 0xFFFFFFFF;

	if (!(entry->flag & SENT_SOFT_SIG)) {
		ret = get_soft_limit(&qentry->entry) -
			get_maze_count(&qentry->entry, task);
	} else {
		ret = get_hard_limit(&qentry->entry) -
			get_maze_count(&qentry->entry, task);
	}

	return ret;
}

/*	
 * Note. Must be called rcu_read_lock()
 */
static int compare_limit(struct maze_queue_entry *new,
						 struct maze_queue_entry *curr)
{
	return get_compare_value(new) - get_compare_value(curr);
}

/*
 * Soft limit excessive check
 * 
 */
static inline int soft_limit(struct maze_entry *entry, 
							 struct task_struct *task, int run_time)
{
	int signal;
	int limit = get_soft_limit(entry);

	if (limit) {
		if (!(entry->sig_flag & SENT_SOFT_SIG) && run_time >= limit) {
			if (entry->soft_sig == MAZE_SIG_USEDEFAULT)
				signal = soft_sig_default;
			else
				signal = entry->soft_sig;

			if (signal > MAZE_SIG_NONE)
				send_sig(signal, task, 1);

			if (use_notice)
				do_notice(task->pid, MAZE_EV_SLIMOVER, limit, 0);

			DBGPRINT("Detect exceeding soft limit"
					 " : PID = %d, lim = %d, sig = %d\n",
					 task->pid, limit, signal);

			entry->sig_flag |= SENT_SOFT_SIG;
			return 1;
		}
	}
	return 0;
}

/*	
 * hard limit excessive check
 * 
 */
static inline int hard_limit(struct maze_entry *entry, 
							 struct task_struct *task, int run_time)
{
	int disp_start, remain;
	int signal;
	int ret = 0;
	int limit = get_hard_limit(entry);

	if (limit) {
		/* Dump registers  */
		if (eip_disp_count > 0) {
			disp_start = limit - 10 * (eip_disp_count - 1);
			remain = limit - run_time;
			/*
			 *  remain < 0 means run_time exceeds hard limit
			 *  After exceeding hard limit,  we continue to
			 *  count up running time, but stop to display EIP.
			 */
			if ((run_time >= disp_start) &&
				(remain >= 0 && remain % 10 == 0)) {
				DBGPRINT("PID %d reaches HARD LIMIT after %d msecs.\n",
						 task->pid, remain);
				ret = 1;
			}
		}

		if (!(entry->sig_flag & SENT_HARD_SIG) && run_time >= limit) {
			if (entry->hard_sig == MAZE_SIG_USEDEFAULT)
				signal = hard_sig_default;
			else
				signal = entry->hard_sig;
			if (signal > MAZE_SIG_NONE)
				send_sig(signal, task, 1);
			if (use_notice)
				do_notice(task->pid, MAZE_EV_HLIMOVER, limit, 0);

			DBGPRINT("Detect exceeding hard limit"
					 " : PID = %d, lim = %d, sig = %d\n",
					 task->pid, limit, signal);

			entry->sig_flag |= SENT_HARD_SIG;
			ret = 1;
		}
	}
	return ret;
}

/*	
 * Note. Must be called rcu_read_lock()
 */
static inline void chk_mazecnt(void)
{
	struct maze_queue_entry *qentry;
	struct maze_queue_itr itr;
	struct maze_entry *entry;
	struct task_struct *task;
	cputime_t count;
	int run_time;

	for_each_queue(&itr, qentry) {
		task = find_task_by_pid(qentry->entry.pid);
		if (task == NULL)
			continue;
		
		might_sleep_list(qentry, task);

		if (task->state != TASK_RUNNING)
			continue;

		count = get_maze_count(&qentry->entry, task);
		run_time = jiffies_to_msecs(count);

		entry = maze_entry(qentry);
		if (!(entry->sig_flag & SENT_SOFT_SIG))
			soft_limit(maze_entry(qentry), task, run_time);
		else
			hard_limit(maze_entry(qentry), task, run_time);
	}
}

/*
 * Soft IRQ timer hander.
 */
static void timer_handler(unsigned long data)
{
	unsigned long long t = begin_timer();
	rcu_read_lock();
	chk_mazecnt();
	rcu_read_unlock();
	continue_timer(timer_handler, maze_timer_interval);
	after_timer(t);
}

static int mzdev_open(struct inode *inode, struct file *file)
{
	atomic_inc(&mzdev_open_count);
	return 0;
}

static int mzdev_release(struct inode *inode, struct file *file)
{
	atomic_inc(&mzdev_release_count);
	return 0;
}

static int ioctl_set_info(unsigned long arg)
{
	struct maze_data md;
	struct task_struct *task, *leader;
	int pid;
	int ret = copy_from_user(&md, (struct maze_data *)arg,
							 sizeof(struct maze_data));
	if (ret) {
		DBGPRINT("MAZE_IOCSETINFO read arg error\n");
		return -EFAULT;
	}

	/* capability check */
	pid = md.pid;
	if (pid && pid != current->pid) {
		if (!capable(CAP_SYS_ADMIN)) {
			DBGPRINT("No permission : PID %d\n",
					 current->pid);
			return -EPERM;
		}
	}

	/* parameter validation check */
	if (pid < 0 || pid == 1)
		return 0;

	/*
	 *  If either soft or hard lims is USEDEFAULT,
	 *  another lim has to be USEDEFAULT.
	 */
	if ((md.soft_lim == MAZE_LIM_USEDEFAULT) ||
		(md.hard_lim == MAZE_LIM_USEDEFAULT)) {
		if (md.soft_lim != md.hard_lim)
			return 0;
	}

	/*
	 *  Otherwise, check magnitude correlation and
	 *  range of both lims.
	 */
	else if ((md.soft_lim < MAZE_LIM_NONE) ||
			 (md.soft_lim > md.hard_lim)   ||
			 (md.hard_lim > MAZE_LIM_MAX))
		return 0;

	/*
	 *  Signal value check
	 */
	if (md.soft_sig < MAZE_SIG_USEDEFAULT ||
		md.soft_sig > SIGRTMAX)
		return 0;

	if (md.hard_sig < MAZE_SIG_USEDEFAULT ||
		md.hard_sig > SIGRTMAX)
		return 0;

	/* set default value */
	if (pid == 0) {
		pid = current->pid;
		md.pid = pid;
	}
	md.flag = md.flag & 0x00000003;

	DBGPRINT("PID:%d  SL:%d  HL:%d	"
			 "SS:%d  HS:%d	FG=%x\n",
			 md.pid, md.soft_lim, md.hard_lim,
			 md.soft_sig, md.hard_sig, md.flag);

	rcu_read_lock();

	task = find_task_by_pid(pid);
	if (!task) {
		DBGPRINT("PID %d not found\n", pid);
		goto rcu_unlock;
	}

	ret = set_maze_entry(&md, task);
	if (ret) {
		DBGPRINT("set_maze_entry for PID %d error\n", pid);
		goto rcu_unlock;
	}

	/*
	 *  thread inherit flag or monitor all flag is on,
	 *  check whether target is a thread group leader.
	 *  If so, set same maze info data for each threads
	 *  in the group.
	 */
	if ((inhr_thread(md.flag) || mon_all()) &&
		thread_group_leader(task) &&
		!thread_group_empty(task)) {
		leader = task;
		for (task = next_thread(leader);
			 task != leader;
			 task = next_thread(task)) {
			md.pid = task->pid;
			ret = set_maze_entry(&md, task);
			if (ret) {
				DBGPRINT("set_maze_entry for PID %d error\n", pid);
				goto rcu_unlock;
			}
		}
	}
rcu_unlock:
	rcu_read_unlock();

	return ret;
}

static int ioctl_get_info(unsigned long arg)
{
	struct maze_data md;
	struct maze_entry *mzent;
	struct task_struct *task;
	int pid;
	int ret = copy_from_user(&md, (struct maze_data *)arg,
							 sizeof(struct maze_data));
	if (ret) {
		DBGPRINT("MAZE_IOCGETINFO : argument error\n");
		return -EFAULT;
	}

	pid = md.pid;
	if (pid < 0 || pid == 1)
		return ret;
	if (pid == 0)
		pid = (int)current->pid;

	rcu_read_lock();

	mzent = find_maze_entry(pid);
	if (unlikely(!mzent)) {
		DBGPRINT("Can't get entry for PID %d\n", pid);
		rcu_read_unlock();
		return -ENODATA;
	}

	task = find_task_by_pid(pid);
	if (unlikely(!task))
		panic("Can't get task_struct for PID %d\n", pid);

	md.pid      = mzent->pid;
	md.soft_lim = mzent->soft_lim;
	md.hard_lim = mzent->hard_lim;
	md.soft_sig = mzent->soft_sig;
	md.hard_sig = mzent->hard_sig;
	md.time     = jiffies_to_msecs(get_maze_count(mzent, task));
	md.flag     = mzent->flag;

	rcu_read_unlock();

	ret = copy_to_user((struct maze_data *)arg, &md,
					   sizeof(struct maze_data));
	if (ret) {
		DBGPRINT("MAZE_IOCGETINFO : argument error\n");
		return -EFAULT;
	}

	return ret;
}

static int ioctl_get_notice(unsigned long arg)
{
	struct maze_notice mn;
	struct notice_entry *nent;
	unsigned long flags;
	int ret;

	DBGPRINT("IOCGETNOTICE start\n");

	if (!capable(CAP_SYS_ADMIN)) {
		DBGPRINT("No permission : PID %d\n", current->pid);
		return -EPERM;
	}

	ret = access_ok(VERIFY_WRITE,
					(void __user *)arg,
					sizeof(struct maze_notice));
	if (ret == 0) {
		DBGPRINT("MAZE_IOCGETNOTICE argument error\n");
		return -EFAULT;
	}

	rcu_read_lock();

	spin_lock_irqsave(&notice_buf_lock, flags);
	nent = get_notice_entry();
	if (!nent) {
		spin_unlock_irqrestore(&notice_buf_lock, flags);
		rcu_read_unlock();
		return -ENODATA;
	}

	mn = nent->mn;
	del_notice_entry(nent);
	spin_unlock_irqrestore(&notice_buf_lock, flags);
	rcu_read_unlock();

	ret = __copy_to_user((struct maze_data *)arg, &mn,
				 sizeof(struct maze_notice));
	if (ret) {
		DBGPRINT("MAZE_IOCGETINFO : argument error\n");
		ret = -EFAULT;
	}

	return ret;
}

static int ioctl_set_attr(unsigned long arg)
{
	struct maze_data md;
	int ret;

	DBGPRINT("IOCSETATTR start\n");

	if (!capable(CAP_SYS_ADMIN)) {
		DBGPRINT("No permission : PID %d\n", current->pid);
		return -EPERM;
	}

	ret = copy_from_user(&md, (struct maze_data *)arg,
						 sizeof(struct maze_data));
	if (ret) {
		DBGPRINT("MAZE_IOCSETATTR argument error\n");
		return -EFAULT;
	}

	/* parameter validation check */
	if (md.soft_lim < MAZE_LIM_NONE ||
		md.soft_lim > md.hard_lim ||
		md.hard_lim > MAZE_LIM_MAX)
		return 0;

	if (md.soft_sig < MAZE_SIG_NONE ||
		md.soft_sig > SIGRTMAX)
		return 0;

	if (md.hard_sig < MAZE_SIG_NONE ||
		md.hard_sig > SIGRTMAX)
		return 0;

	/* set default value */
	soft_lim_default = md.soft_lim;
	hard_lim_default = md.hard_lim;
	soft_sig_default = md.soft_sig;
	hard_sig_default = md.hard_sig;
	maze_attr_flag = md.flag & 0x00000007;

	DBGPRINT("DEFAULT SL:%d  HL:%d "
			 "SS:%d  HS:%d  FG=%x\n",
			 soft_lim_default, hard_lim_default,
			 soft_sig_default, hard_sig_default,
			 maze_attr_flag);
	return 0;
}

static int mzdev_ioctl(struct inode *inode, struct file *file,
					   unsigned int cmd, unsigned long arg)
{
	int ret = -EINVAL;

	if ((_IOC_TYPE(cmd) != MAZE_IOC_MAGIC) ||
	    (_IOC_NR(cmd) > MAZE_IOC_MAXNR)) {
		ERRPRINT("Invalid ioctl command : %x\n", cmd);
		return ret;
	}

	switch (cmd) {
	case MAZE_IOCSETINFO:
	  ret = ioctl_set_info(arg);
	  break;
	case MAZE_IOCGETINFO:
	  ret = ioctl_get_info(arg);
	  break;
	case MAZE_IOCGETNOTICE:
	  ret = ioctl_get_notice(arg);
	  break;
	case MAZE_IOCSETATTR:
	  ret = ioctl_set_attr(arg);
	break;
	}

	return ret;
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 11))
static long mzdev_compat_ioctl(struct file *file, unsigned cmd,
							   unsigned long arg)
{
    return mzdev_ioctl(file->f_dentry->d_inode, file, cmd, arg);
}
#endif

static unsigned int mzdev_poll (struct file *file, poll_table *wait)
{
	unsigned int mask = 0;

	if (list_empty(&notice_buf_list))
		poll_wait(file, &maze_wait, wait);

	if (!(list_empty(&notice_buf_list))) {
		mask = POLLIN | POLLRDNORM;
		DBGPRINT("set return mask\n");
	}

	return mask;
}

static struct file_operations maze_fops = {
	.open	 = mzdev_open,
	.release = mzdev_release,
	.ioctl	 = mzdev_ioctl,
	.poll	 = mzdev_poll,
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 11))
	.compat_ioctl = mzdev_compat_ioctl,
#endif
};

static struct class maze_class = {
	.name  = MZCLASS_NAME,
	.owner = THIS_MODULE,
};

static int __init init_devno(void)
{
	int ret;

	if (mazedev_major) {
		maze_devno = MKDEV(mazedev_major, MZDEV_MINOR);
		ret = register_chrdev_region(maze_devno, MZDEV_NUM, MZDEV_NAME);
		if (unlikely(ret)) {
			ERRPRINT("register_chrdev_region() failed,"
					 "rc=%d, major=%d\n", ret, mazedev_major);
			return ret;
		}
	} else {
		ret = alloc_chrdev_region(&maze_devno, MZDEV_MINOR,
								  MZDEV_NUM, MZDEV_NAME);
		if (unlikely(ret)) {
			ERRPRINT("alloc_chrdev_region() failed, rc=%d\n", ret);
			return ret;
		}
		mazedev_major = MAJOR(maze_devno);
	}
	DBGPRINT("maze driver register OK!"
			 " : major = %d\n", mazedev_major);

	return 0;
}

static int __init init_cdev(void)
{
	int ret;

	maze_cdev = cdev_alloc();
	maze_cdev->owner = THIS_MODULE;
	maze_cdev->ops = &maze_fops;
	ret = cdev_add(maze_cdev, maze_devno, MZDEV_NUM);
	if (unlikely(ret))
		return ret;

	return 0;
}

static int __init init_class(void)
{
	int ret = class_register(&maze_class);

	if (unlikely(ret)) {
		ERRPRINT("fail to regist class\n");
		return ret;
	}

	return 0;
}

static int __init init_class_device(void)
{
	int ret;
	struct class_device *class_dev =
		class_device_create(&maze_class, NULL, maze_devno,
							NULL, MZCLASS_NAME);

	if (IS_ERR(class_dev)) {
		ERRPRINT("fail to create class_device\n");
		ret = PTR_ERR(class_dev);
		return ret;
	}

	return 0;
}

/*
 * init maze hash spin_lock
 */
static void __init init_maze_hash(void)
{
	int i;
	for (i = 0; i < MZHASH_NR; i++) {
		INIT_HLIST_HEAD(&mz_hash[i]);
		spin_lock_init(&mzhlist_lock[i]);
	}
}

static struct maze_queue_ops queue_ops = {
	.compare = compare_limit,
};

/*
 *  Initialization of maze driver
 */
static int __init init_mzdriver(void)
{
	int ret;

	atomic_set(&mzdev_open_count, 0);
	atomic_set(&mzdev_release_count, 0);

#ifdef CONFIG_MAZEPROC
	if (mzdev_install_proc())
		return -EIO;
#endif
	ret = init_devno();
	if (ret)
		goto remove_proc;

	ret = init_cdev();
	if (ret)
		goto remove_chrdev_region;

	ret = init_class();
	if (ret)
		goto remove_cdev;

	ret = init_class_device();
	if (ret)
		goto remove_class;

	init_queue(&queue_ops);
	init_maze_hash();

	/* create notice entry cache */
	spin_lock_init(&notice_buf_lock);

	/* set timer to check maze count  */
	set_cpu_timer(timer_handler, maze_timer_interval);

	DBGPRINT("maze driver initialized Okay!\n");
	DBGPRINT("eip_disp_count   = %d\n", eip_disp_count);
	DBGPRINT("use_notice       = %d\n", use_notice);
	DBGPRINT("notice_buf_count = %d\n", notice_buf_count);
	DBGPRINT("mazedev_major    = %d\n", mazedev_major);

	return 0;

remove_class:
	class_unregister(&maze_class);
remove_cdev:
	cdev_del(maze_cdev);
remove_chrdev_region:
	unregister_chrdev_region(maze_devno, MZDEV_NUM);
remove_proc:
	mzdev_remove_proc();

	return ret;
}

/*
 * detach all utrace engine.
 */
static void __exit detach_all_utrace(void)
{
	struct utrace_attached_engine *engine;
	struct maze_entry *entry;
	struct task_struct *task;
	struct hlist_node *node;
	unsigned long flags;
	int i;

	for (i = 0; i < MZHASH_NR; i++) {
		spin_lock_irqsave(&mzhlist_lock[i], flags);
		hlist_for_each_entry(entry, node, &mz_hash[i], chain) {
			rcu_read_lock();
			task = find_task_by_pid(entry->pid);
			if (task) {
				 engine =
					 utrace_attach(task, UTRACE_ATTACH_MATCH_OPS,
								   &task_ops, 0);
				 if (!IS_ERR_VALUE((long)engine)) {
					 utrace_detach(task, engine);
				 } else {
					 DBGPRINT("fail to detach utrace."
							  "PID = %d\n",
							  entry->pid);
				 }
			}
			rcu_read_unlock();
		}
		spin_unlock_irqrestore(&mzhlist_lock[i], flags);
	}
}

/*
 * remove all nodes in the maze hash table
 */
static void __exit remove_all_mz_hash(void)
{
	struct maze_entry *entry;
	unsigned long flags;
	int i;

	lock_all_queue();

	for (i = 0; i < MZHASH_NR; i++) {
		spin_lock_irqsave(&mzhlist_lock[i], flags);
		for (;;) {
			if (hlist_empty(&mz_hash[i]))
				break;
			entry =
				hlist_entry(mz_hash[i].first, struct maze_entry, chain);
			hlist_del_rcu(&entry->chain);
			synchronize_rcu();
			free_queue_entry(maze_queue_entry(entry));
		}
		spin_unlock_irqrestore(&mzhlist_lock[i], flags);
	}	

	unlock_all_queue();

	DBGPRINT("Remove all maze hash\n");
}

/*
 * remove all nodes in the notice buffer list
 */
static void __exit remove_all_notice_buf_list(void)
{
	struct notice_entry *entry;
	struct notice_entry *next;

	if (!list_empty(&notice_buf_list)) {
		list_for_each_entry_safe(entry, next, &notice_buf_list, chain) {
			DBGPRINT("delete event for pid = %d\n", entry->mn.pid);
			list_del(&(entry->chain));
			kfree(entry);
		}
	}
}

/*
 * Exit maze driver
 */
static void __exit exit_mzdriver(void)
{
	unsigned long flags;

	/* delete timer to check maze count */
	del_all_cpu_timer_sync();

	/* unregister device from kernel */
	cdev_del(maze_cdev);
	unregister_chrdev_region(maze_devno, MZDEV_NUM);

	/* detach all utrace engine */
	detach_all_utrace();

#ifdef CONFIG_MAZEPROC
	/* remove /proc entry */
	mzdev_remove_proc();
	DBGPRINT("Remove proc files\n");
#endif

	/* destroy maze class_device */
	class_device_destroy(&maze_class, maze_devno);

	/* unregist maze class */
	class_unregister(&maze_class);

	remove_all_mz_hash();

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 18))
	rcu_read_lock();
	rcu_barrier();
	rcu_read_unlock();
#endif

	spin_lock_irqsave(&notice_buf_lock, flags);
	remove_all_notice_buf_list();
	spin_unlock_irqrestore(&notice_buf_lock, flags);

	DBGPRINT("maze driver ready for removal\n");
}

module_init(init_mzdriver);
module_exit(exit_mzdriver);

MODULE_AUTHOR("MIRACLE LINUX");
MODULE_DESCRIPTION("MAZE monitorgring driver");
MODULE_LICENSE("GPL");
