/*
 *  Copyright (C) 2007-2008 MIRACLE LINUX Corp.
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 */

#include <linux/module.h>	/* Definitions needed for kernel modules */
#include <linux/init.h>
#include <linux/kernel.h>	/* We run in the kernel so we need this */
#include <linux/proc_fs.h>	/* The /proc definitions are in this one */
#include <linux/seq_file.h>
#include <linux/hash.h>
#include <linux/time.h>
#include <linux/uaccess.h>
#include <linux/types.h>

#include "maze.h"
#include "mazequeue.h"
#include "mazetimer.h"

extern struct hlist_head mz_hash[];

extern struct list_head notice_buf_list;
extern int notice_num;

/*
 *  Default value
 */
extern int soft_lim_default;
extern int hard_lim_default;
extern int soft_sig_default;
extern int hard_sig_default;
extern int maze_attr_flag;

/*
 *  Module parameters
 */
extern int eip_disp_count;
extern int use_notice;
extern int notice_buf_count;
extern int mazedev_major;

/*
 *  Prototypes for functions
 */
extern int mzdev_install_proc(void);
extern void mzdev_remove_proc(void);

/*
 * data for device driver
 */
static struct proc_dir_entry *mz_proc_root;

/*
 *  Seq file operations for /proc/maze/ent
 */
static void *procmz_ent_seq_start(struct seq_file *s, loff_t *pos)
{
	/* read_lock(&tasklist_lock); */

	while (*pos < MZHASH_NR) {
		if (!hlist_empty(&mz_hash[*pos]))
			return &mz_hash[(*pos)++];
		(*pos)++;
	}
	return NULL;
}

static void *procmz_ent_seq_next(struct seq_file *s, void *v, loff_t *pos)
{
	while (*pos < MZHASH_NR) {
		if (!hlist_empty(&mz_hash[*pos]))
			return &mz_hash[(*pos)++];
		(*pos)++;
	}

	return NULL;
}

static void procmz_ent_seq_stop(struct seq_file *s, void *v)
{
	/* read_unlock(&tasklist_lock); */
}

static int procmz_ent_seq_show(struct seq_file *s, void *v)
{
	struct hlist_head *mz_head = (struct hlist_head *)v;
	struct hlist_node *elem;
	struct maze_entry *p;
	struct task_struct *task;

	rcu_read_lock();
	hlist_for_each_entry_rcu(p, elem, mz_head, chain) {
		task = find_task_by_pid(p->pid);
		if (unlikely(!task))
			panic("Can't get task_struct for PID %d\n", p->pid);

		seq_printf(s, "PID:%5d   ", p->pid);
		seq_printf(s, "slim:%6d  ", p->soft_lim);
		seq_printf(s, "hlim:%6d  ", p->hard_lim);
		seq_printf(s, "ssig:%2d  ", p->soft_sig);
		seq_printf(s, "hsig:%2d  ", p->hard_sig);
		seq_printf(s, "cnt:%5d   ", jiffies_to_msecs(get_maze_count(p, task)));
		seq_printf(s, "flag:0x%x", p->flag);
		seq_printf(s, "\n");
	}
	rcu_read_unlock();

	return 0;
}

static struct seq_operations procmz_ent_seq_ops = {

	.start	= procmz_ent_seq_start,
	.next	= procmz_ent_seq_next,
	.stop	= procmz_ent_seq_stop,
	.show	= procmz_ent_seq_show
};

static int procmz_ent_seq_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &procmz_ent_seq_ops);
}

static struct file_operations procmz_ent_ops = {
	.owner	 = THIS_MODULE,
	.open	 = procmz_ent_seq_open,
	.read	 = seq_read,
	.llseek	 = seq_lseek,
	.release = seq_release
};

/*
 *  Seq file operation methods for /dev/maze/data
 */
static int procmz_data_seq_show(struct seq_file *s, void *v)
{
	seq_printf(s, "Default Data Settings:\n");
	seq_printf(s, "  Soft Limit     : %d\n", soft_lim_default);
	seq_printf(s, "  Hard Limit     : %d\n", hard_lim_default);
	seq_printf(s, "  Soft Signal    : %d\n", soft_sig_default);
	seq_printf(s, "  Hard Signal    : %d\n", hard_sig_default);
	seq_printf(s, "  Attribute      : 0x%x\n", maze_attr_flag);

	seq_printf(s, "\n");

	seq_printf(s, "Module Parameters:\n");
	seq_printf(s, "  eip_disp_count   : %d\n", eip_disp_count);
	seq_printf(s, "  use_notice       : %d\n", use_notice);
	seq_printf(s, "  notice_buf_count : %d\n", notice_buf_count);
	seq_printf(s, "  mazedev_major    : %d\n", mazedev_major);

	seq_printf(s, "\n");

	return 0;
}

static int procmz_data_seq_open(struct inode *inode, struct file *file)
{
	return single_open(file, procmz_data_seq_show, NULL);
}

static struct file_operations procmz_data_ops = {
	.owner	 = THIS_MODULE,
	.open	 = procmz_data_seq_open,
	.read	 = seq_read,
	.llseek	 = seq_lseek,
	.release = seq_release
};

#ifdef CONFIG_MAZEPROCDEBUG

/*
 *  Seq file operation methods for /dev/maze/debug
 */
static int procmz_debug_seq_show(struct seq_file *s, void *v)
{
	int i;

	for_each_online_cpu(i) {
		seq_printf(s, "CPU%d:%lld\n", i, get_ave_time(i));
	}

/* 	seq_printf(s, "cpu  queue  sleep   timer   peek     ave\n"); */
/* 	for_each_possible_cpu(i) { */
/* 		seq_printf(s, " %d     %d      %d       %d       %lld     %lld\n", i, */
/* 				   get_amount_queue(i), */
/* 				   get_sleeping_list_len(i), */
/* 				   is_active_timer(i), */
/* 				   get_peek_time(i), */
/* 				   get_ave_time(i)); */
/* 	} */

	return 0;
}

static int procmz_debug_seq_open(struct inode *inode, struct file *file)
{
	return single_open(file, procmz_debug_seq_show, NULL);
}

static struct file_operations procmz_debug_ops = {
	.owner	 = THIS_MODULE,
	.open	 = procmz_debug_seq_open,
	.read	 = seq_read,
	.llseek	 = seq_lseek,
	.release = seq_release
};

static int procmz_queue_debug_seq_open(struct inode *inode, struct file *file)
{
	return single_open(file, procmz_queue_debug_seq_show, NULL);
}

static struct file_operations procmz_queue_debug_ops = {
	.owner	 = THIS_MODULE,
	.open	 = procmz_queue_debug_seq_open,
	.read	 = seq_read,
	.llseek	 = seq_lseek,
	.release = seq_release
};

static inline int create_maze_debug(void)
{
	struct proc_dir_entry *p;

	p = create_proc_entry("debug", S_IRUGO, mz_proc_root);
	if (p == NULL) {
		ERRPRINT("/proc/maze/debug create error\n");
		return  -1;
	}
	p->proc_fops = &procmz_debug_ops;

	p = create_proc_entry("queue_debug", S_IRUGO, mz_proc_root);
	if (p == NULL) {
		ERRPRINT("/proc/maze/queue_debug create error\n");
		return  -1;
	}
	p->proc_fops = &procmz_queue_debug_ops;


	return 0;
}

static inline void remove_maze_debug(void)
{
	remove_proc_entry("debug", mz_proc_root);
	remove_proc_entry("queue_debug", mz_proc_root);
	
}

#else

static inline int create_maze_debug(void)
{
	return 0;
}

static inline void remove_maze_debug(void)
{
}
#endif

static inline int create_maze_dir(void)
{
	/* root node  */
	mz_proc_root =
		create_proc_entry(MZDEV_NAME,
						  S_IFDIR | S_IRUGO | S_IXUGO,
						  NULL);

	if (mz_proc_root == NULL) {
		ERRPRINT("/proc/maze create error\n");
		return -1;
	}

	return 0;
}

static inline int create_maze_ent(void)
{
	struct proc_dir_entry *p;

	/* ent node */
	p = create_proc_entry("ent", S_IRUGO, mz_proc_root);
	if (p == NULL) {
		ERRPRINT("/proc/maze/ent create error\n");
		return  -1;
	}
	p->proc_fops = &procmz_ent_ops;

	return 0;
}

static inline int create_maze_data(void)
{
	struct proc_dir_entry *p;

	/* data node */
	p = create_proc_entry("data", S_IRUGO, mz_proc_root);
	if (p == NULL) {
		ERRPRINT("/proc/maze/data create error\n");
		return  -1;
	}
	p->proc_fops = &procmz_data_ops;

	return 0;
}

/*
 *  Install the driver's /proc handler
 */
int mzdev_install_proc(void)
{
	int ret;

	ret = create_maze_dir();
	if (ret < 0)
		goto err_ret;

	ret = create_maze_ent();
	if (ret < 0)
		goto err_ret;

	ret = create_maze_data();
	if (ret < 0)
		goto err_ret;

	ret = create_maze_debug();
	if (ret < 0)
		goto err_ret;

	return 0;

err_ret:

	return -1;
}

/*
 *  Remove the driver's /proc handler
 */
void mzdev_remove_proc(void)
{
	remove_proc_entry("ent", mz_proc_root);
	remove_proc_entry("data", mz_proc_root);
	remove_maze_debug();
	remove_proc_entry(MZDEV_NAME, NULL);
}
