/*
 *  mazeque_hash.c
 *
 *  Copyright (C) 2007-2008 MIRACLE LINUX Corp.
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 */
#include <linux/cpumask.h>
#include <linux/list.h>
#include <linux/percpu.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/slab.h>
#include <linux/jiffies.h>
#include "maze.h"
#include "mazequeue.h"

#define QUEUE_HASH_NR 25
#define QUEUE_LOW_NR 128

struct maze_queue
{
	struct hlist_head hash[QUEUE_HASH_NR];
	unsigned long last[QUEUE_HASH_NR];
	spinlock_t lock;
	struct maze_queue_ops *ops;
};

DEFINE_PER_CPU(struct maze_queue, maze_queue);
DEFINE_PER_CPU(unsigned long, maze_queue_lock_flags);

#ifdef CONFIG_MAZEPROCDEBUG
DEFINE_PER_CPU(unsigned long, hash_amount[QUEUE_HASH_NR]);

static inline void inc_hash_amount(struct maze_queue_entry *qentry, int nr)
{
	qentry->nr = nr;
	__get_cpu_var(hash_amount)[nr]++;
}

static inline void dec_hash_amount(struct maze_queue_entry *qentry)
{
	__get_cpu_var(hash_amount)[qentry->nr]--;
}
#else

static inline void inc_hash_amount(struct maze_queue_entry *qentry, int nr)
{
}

static inline void dec_hash_amount(struct maze_queue_entry *qentry)
{
}
#endif

static inline int queue_hashfn(struct maze_queue_entry *qentry)
{
	struct task_struct *task;
	cputime_t t;
	int i;

	task = find_task_by_pid(maze_entry(qentry)->pid);
	if (unlikely(!task))
		goto max_nr;

	t = get_maze_count(maze_entry(qentry), task);
	if (maze_entry(qentry)->sig_flag & SENT_SOFT_SIG)
		t -= maze_entry(qentry)->hard_lim;
	else
		t -= maze_entry(qentry)->soft_lim;

	for (i = 0;i < QUEUE_HASH_NR;i++) {
		if (QUEUE_LOW_NR << i > t)
			return i;
	}

max_nr:
	return QUEUE_HASH_NR - 1;
}

static inline void __init_queue(int cpu, struct maze_queue_ops *not_use)
{
	int i;
	unsigned long t = jiffies;

	for (i = 0;i < QUEUE_HASH_NR; i++) {
		INIT_HLIST_HEAD(&per_cpu(maze_queue, cpu).hash[i]);
		per_cpu(maze_queue, cpu).last[i] = t;
	}
	spin_lock_init(&per_cpu(maze_queue, cpu).lock);
}

static inline void insert_entry_on(struct maze_queue_entry *qentry, int cpu)
{
	int nr = queue_hashfn(qentry);
	hlist_add_head(&qentry->chain, &per_cpu(maze_queue, cpu).hash[nr]);
	qentry->state |= ENQUEUE_MAZE;
	inc_hash_amount(qentry, nr);
}

static inline void remove_entry_on(struct maze_queue_entry *qentry, int cpu)
{
	if (!(qentry->state & ENQUEUE_MAZE))
		return;
	hlist_del(&qentry->chain);
	dec_hash_amount(qentry);
}

void fastcall __init_queue_itr(struct maze_queue_itr *itr)
{
	itr->queue = &__get_cpu_var(maze_queue);
	itr->curr = itr->queue->hash[0].first;
	itr->queue->last[0] = jiffies;
	itr->hash_nr = 0;
}

struct maze_queue_entry* fastcall next_queue_entry(struct maze_queue_itr *itr)
{
	struct maze_queue_entry *qentry;
	unsigned long m;
	int nr;

restart:
	while (!itr->curr) {
		itr->hash_nr++;
		if (itr->hash_nr >= QUEUE_HASH_NR)
			return NULL;

		m = jiffies_to_msecs(jiffies - itr->queue->last[itr->hash_nr]);
		if (m < QUEUE_LOW_NR >> itr->hash_nr)
			return NULL;
		itr->queue->last[itr->hash_nr] = jiffies;
		itr->curr = itr->queue->hash[itr->hash_nr].first;
	}

	qentry = hlist_entry(itr->curr, struct maze_queue_entry, chain);	
	itr->curr = itr->curr->next;

	if (itr->hash_nr) {
		nr = queue_hashfn(qentry);
		if (nr < itr->hash_nr) {
			hlist_del(&qentry->chain);
			hlist_add_head(&qentry->chain, &__get_cpu_var(maze_queue).hash[nr]);

			dec_hash_amount(qentry);
			inc_hash_amount(qentry, nr);
		} else if (nr > itr->hash_nr) {
			hlist_del(&qentry->chain);
			hlist_add_head(&qentry->chain, &__get_cpu_var(maze_queue).hash[nr]);
			qentry = NULL;

			dec_hash_amount(qentry);
			inc_hash_amount(qentry, nr);
		}	
	}

	if (!qentry)
		goto restart;
	
	return qentry;
}

#ifdef CONFIG_MAZEPROCDEBUG
int procmz_queue_debug_seq_show(struct seq_file *s, void *v)
{
	int i;
	int nr;

	for_each_online_cpu(i) {
		seq_printf(s, "cpu %d\n", i);
		for (nr = 0;nr < QUEUE_HASH_NR; nr++)
			seq_printf(s, "%d,%ld\n", nr, per_cpu(hash_amount, i)[nr]);
		seq_printf(s, "\n");
	}

	return 0;
}
#endif

#include "queue_internal.c"
