/*
 *  queue_internal.c
 *
 *  Copyright (C) 2007-2008 MIRACLE LINUX Corp.
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 */
#include <linux/cpumask.h>
#include <linux/list.h>
#include <linux/percpu.h>
#include <linux/slab.h>

DEFINE_PER_CPU(struct list_head, operation_list);
DEFINE_PER_CPU(spinlock_t, operation_list_lock);

DEFINE_PER_CPU(atomic_t, maze_queue_amount);
static spinlock_t maze_queue_amount_lock;

DEFINE_PER_CPU(struct list_head, sleeping_list);
DEFINE_PER_CPU(spinlock_t, sleeping_list_lock);

static inline void init_amount(int cpu)
{
	atomic_set(&per_cpu(maze_queue_amount, cpu), 0);
	DBGPRINT("initialize cpu %d : maze_queue_amount = %d\n",
			 cpu, atomic_read(&per_cpu(maze_queue_amount, cpu)));
}
static inline void inc_amount(int cpu)
{
	atomic_inc(&per_cpu(maze_queue_amount, cpu));
	DBGPRINT("increment cpu %d : maze_queue_amount = %d\n",
			 cpu, atomic_read(&per_cpu(maze_queue_amount, cpu)));
}
static inline void dec_amount(int cpu)
{
	atomic_inc(&per_cpu(maze_queue_amount, cpu));
	DBGPRINT("decrement cpu %d : maze_queue_amount = %d\n",
			 cpu, atomic_read(&per_cpu(maze_queue_amount, cpu)));
}

static void insert_entry(struct maze_queue_entry *qentry)
{
	insert_entry_on(qentry, qentry->cpu);
	inc_amount(qentry->cpu);
}

static void remove_entry(struct maze_queue_entry *qentry)
{
	remove_entry_on(qentry, qentry->cpu);
	dec_amount(qentry->cpu);
}

void init_queue(struct maze_queue_ops *ops)
{
	int i;
	for_each_possible_cpu(i) {
		spin_lock_init(&per_cpu(operation_list_lock, i));
		INIT_LIST_HEAD(&per_cpu(operation_list, i));
		spin_lock_init(&per_cpu(sleeping_list_lock, i));
		INIT_LIST_HEAD(&per_cpu(sleeping_list, i));
		__init_queue(i, ops);
		init_amount(i);
	}
	spin_lock_init(&maze_queue_amount_lock);
}

static inline void add_operation_list(struct maze_queue_entry *qentry,
										 unsigned long cmd)
{
	unsigned long flags;
	unsigned long tmp;

	DBGPRINT("add command = %ld\n", cmd);

	spin_lock_irqsave(&per_cpu(operation_list_lock, qentry->cpu), flags);

	tmp = qentry->command;
	qentry->command |= cmd;

	if (!tmp)
		list_add_tail(&qentry->clist, &per_cpu(operation_list, qentry->cpu));

	spin_unlock_irqrestore(&per_cpu(operation_list_lock, qentry->cpu), flags);
}

static inline int rr_maze_queue(void)
{
	int ret,i;
	unsigned long flags,low,t;

	ret = 0;
	low = ~0;

	spin_lock_irqsave(&maze_queue_amount_lock, flags);
	for_each_online_cpu(i) {
		t = atomic_read(&per_cpu(maze_queue_amount, i));
		if (low > t) {
			ret = i;
			low = t;
		}
	}
	spin_unlock_irqrestore(&maze_queue_amount_lock, flags);

	return ret;
}

void enqueue(struct maze_queue_entry *qentry)
{
	qentry->cpu = rr_maze_queue();
	add_operation_list(qentry, ENQUEUE_MAZE);
}

void dequeue(struct maze_queue_entry *qentry)
{
	add_operation_list(qentry, DEQUEUE_MAZE);
}

void dequeue_and_free(struct maze_queue_entry *qentry)
{
	add_operation_list(qentry, DEQUEUE_AND_FREE_MAZE);
}

void migration(struct maze_queue_entry *qentry)
{
	//add_operation_list(qentry, MIGRATION_MAZE);
}

void rebalance(struct maze_queue_entry *qentry)
{
	add_operation_list(qentry, REBALANCE_MAZE);
}

void might_sleep_list(struct maze_queue_entry *qentry, struct task_struct *task)
{
	if (task->state == TASK_RUNNING) {
		if (qentry->in_slist)
			add_operation_list(qentry, WAKEUP_MAZE);
	} else {
		if (!qentry->in_slist)
			add_operation_list(qentry, INTO_SLEEPING_MAZE);
	}
}

static void free_callback(struct rcu_head *head)
{
	struct maze_entry *entry = container_of(head, struct maze_entry, rcu_head);
	struct maze_queue_entry *qentry = maze_queue_entry(entry);
	free_queue_entry(qentry);
}

static void insert_sleeping_list(struct maze_queue_entry *qentry)
{
	unsigned long flags;
	spin_lock_irqsave(&__get_cpu_var(sleeping_list_lock), flags);
	list_add_tail(&qentry->slist, &__get_cpu_var(sleeping_list));
	qentry->in_slist = 1;
	spin_unlock_irqrestore(&__get_cpu_var(sleeping_list_lock), flags);
}

static void remove_sleeping_list(struct maze_queue_entry *qentry)
{
	unsigned long flags;
	spin_lock_irqsave(&__get_cpu_var(sleeping_list_lock), flags);
	list_del(&qentry->slist);
	qentry->in_slist = 0;
	spin_unlock_irqrestore(&__get_cpu_var(sleeping_list_lock), flags);
}

/*
 * 
 * Note. Must be called rcu_read_lock().
 */
void __exec_cmds(void)
{
	struct list_head head;
	struct maze_queue_entry *qentry;
	struct maze_queue_entry *next;
	unsigned long flags;

	spin_lock_irqsave(&__get_cpu_var(operation_list_lock), flags);
	list_replace(&__get_cpu_var(operation_list), &head);
	INIT_LIST_HEAD(&__get_cpu_var(operation_list));
	spin_unlock_irqrestore(&__get_cpu_var(operation_list_lock), flags);

	list_for_each_entry_safe(qentry, next, &head, clist) {
		spin_lock_irqsave(&__get_cpu_var(maze_queue).lock, flags);

		if (unlikely(qentry->cpu != smp_processor_id()))
			panic("Don't mach execute cpu");

		if (unlikely(!qentry->command))
			ERRPRINT("Command of maze_queue_entry is zero.\n");

		if ((qentry->command & DEQUEUE_AND_FREE_MAZE) == DEQUEUE_AND_FREE_MAZE) {
			if (qentry->in_slist)
				remove_sleeping_list(qentry);
			else
				remove_entry(qentry);
			qentry->command = 0;
			call_rcu(&maze_entry(qentry)->rcu_head, free_callback);
		}
		if ((qentry->command & NOTHING_MAZE) == NOTHING_MAZE) {
			qentry->command &= ~NOTHING_MAZE;
		}
		if ((qentry->command & WAKEUP_MAZE) == WAKEUP_MAZE) {
			remove_sleeping_list(qentry);
			insert_entry(qentry);
			qentry->command &= ~WAKEUP_MAZE;
		}
		if ((qentry->command & INTO_SLEEPING_MAZE) == INTO_SLEEPING_MAZE) {
			remove_entry(qentry);
			insert_sleeping_list(qentry);
			qentry->command &= ~INTO_SLEEPING_MAZE;
		}
		if ((qentry->command & MIGRATION_MAZE) == MIGRATION_MAZE) {
			//XXX Don't use
			qentry->command &= ~MIGRATION_MAZE;
		}
		if ((qentry->command & REBALANCE_MAZE) == REBALANCE_MAZE) {
			remove_entry(qentry);
			insert_entry(qentry);
			qentry->command &= ~REBALANCE_MAZE;
		}
		if ((qentry->command & ENQUEUE_MAZE) == ENQUEUE_MAZE) {
			insert_entry(qentry);
			qentry->command &= ~ENQUEUE_MAZE;
		}
		if ((qentry->command & DEQUEUE_MAZE) == DEQUEUE_MAZE) {
			if (qentry->in_slist)
				remove_sleeping_list(qentry);
			else
				remove_entry(qentry);
			qentry->command &= ~DEQUEUE_MAZE;
		}

		list_del(&qentry->clist);
		spin_unlock_irqrestore(&__get_cpu_var(maze_queue).lock, flags);
	}
}

void lock_all_queue(void)
{
	int i;
	for_each_possible_cpu(i) {
		spin_lock_irqsave(&per_cpu(maze_queue, i).lock,
						  per_cpu(maze_queue_lock_flags, i));
	}
}

void unlock_all_queue(void)
{
	int i;
	for_each_possible_cpu(i) {
		spin_unlock_irqrestore(&per_cpu(maze_queue, i).lock,
							   per_cpu(maze_queue_lock_flags, i));
	}
}

int get_amount_queue(int cpu)
{
	return atomic_read(&per_cpu(maze_queue_amount, cpu));
}

int get_sleeping_list_len(int cpu)
{
	struct list_head *pos;
	unsigned long flags;
	int i = 0;

	spin_lock_irqsave(&per_cpu(sleeping_list_lock, cpu), flags);
	list_for_each(pos, &per_cpu(sleeping_list, cpu)) {
		i++;
	}
	spin_unlock_irqrestore(&per_cpu(sleeping_list_lock, cpu), flags);

	return i;
}
